#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from __future__ import with_statement

import os
import re
import codecs
import logging.handlers
import time

from taurus.core.taurushelper import Device, Factory
from taurus.core.taurusbasetypes import TaurusEventType
from taurus.core.util.containers import CaselessDict
from taurus.core.util.threadpool import ThreadPool
from taurus.core.util.log import Logger
from taurus.core.tango.sardana.motion import Motion, MotionGroup
from taurus.core.tango.sardana.pool import registerExtensions

from sardana import InvalidId, ElementType, Interface
from sardana.sardanaevent import EventType
from sardana.sardanamodulemanager import ModuleManager
from sardana.sardanamanager import SardanaElementManager, SardanaIDManager
from sardana.sardanathreadpool import get_thread_pool
from sardana.sardanautils import is_pure_str

from .msbase import MSObject
from .mscontainer import MSContainer
from .msdoor import MSDoor
from .msmacromanager import MacroManager
from .mstypemanager import TypeManager
from .msenvmanager import EnvironmentManager
from .msparameter import ParamType
from .msexception import UnknownMacroLibrary, UnknownMacro

CHANGE_EVT_TYPES = TaurusEventType.Change, TaurusEventType.Periodic

ET = ElementType
#: dictionary dict<:data:`~sardana.ElementType`, :class:`tuple`> 
#: where tuple is a sequence:
#: 
#: #. type string representation
#: #. family
#: #. internal macro server class
#: #. automatic full name
TYPE_MAP = {
    ET.Door : ("Door", "Door", MSDoor, "door/{macro_server.name}/{name}"),
}

class TypeData(object):
    """Information for a specific Element type"""
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

#: dictionary
#: dict<:data:`~sardana.ElementType`, :class:`~sardana.macroserver.macroserver.TypeData`>
TYPE_MAP_OBJ = {}
for t, d in TYPE_MAP.items():
    o = TypeData(type=t, name=d[0], family=d[1], klass=d[2] ,
                 auto_full_name=d[3])
    TYPE_MAP_OBJ[t] = o


class NonOverlappingTimedRotatingFileHandler(logging.handlers.TimedRotatingFileHandler):

    def getNewFileName(self):
        currentTime = int(time.time())
        dstNow = time.localtime(currentTime)[-1]
        t = self.rolloverAt - self.interval
        if self.utc:
            timeTuple = time.gmtime(t)
        else:
            timeTuple = time.localtime(t)
            dstThen = timeTuple[-1]
            if dstNow != dstThen:
                if dstNow:
                    addend = 3600
                else:
                    addend = -3600
                timeTuple = time.localtime(t + addend)
        
        dfn = self.baseFilename + "." + time.strftime(self.suffix, timeTuple)
        # PY3K
        if hasattr(self, 'rotation_filename'):
            dfn = self.rotation_filename(dfn)
        return dfn

    def doRollover(self):
        dfn = self.getNewFileName()
        do_backup = os.path.isfile(dfn) and self.backupCount > 0
        if do_backup:
            dfn_backup = dfn + os.path.extsep + "backup"
            os.rename(dfn, dfn_backup)
        try:
            super(NonOverlappingTimedRotatingFileHandler, self).doRollover()
        finally:
            if do_backup:
                mode = self.mode
                if 'a' not in mode:
                    mode = 'a' + mode
                with codecs.open(dfn_backup, mode, self.encoding) as dest:
                    with codecs.open(dfn, 'r', self.encoding) as src:
                        dest.write(src.read())
                os.remove(dfn)
                os.rename(dfn_backup, dfn)


class MacroServer(MSContainer, MSObject, SardanaElementManager, SardanaIDManager):
    
    All = "All"
    
    MaxParalellMacros = 5
    
    logReportParams = dict(when='midnight', interval=1, backupCount=365)
    logReportKlass = NonOverlappingTimedRotatingFileHandler
    
    def __init__(self, full_name, name=None, macro_path=None,
                 environment_db=None):
        # dict<str, Pool>
        # key   - device name (case insensitive)
        # value - Pool object representing the device name
        self._pools = CaselessDict()
        self._max_parallel_macros = self.MaxParalellMacros
        self._path_id = None
        
        MSContainer.__init__(self)
        MSObject.__init__(self, full_name=full_name, name=name, id=InvalidId,
                          macro_server=self, elem_type=ElementType.MacroServer)
        
        registerExtensions()
        
        self._type_manager = TypeManager(self)
        self._environment_manager = EnvironmentManager(self,
                                        environment_db=environment_db)
        self._macro_manager = MacroManager(self, macro_path=macro_path)

    def serialize(self, *args, **kwargs):
        kwargs = MSObject.serialize(self, *args, **kwargs)
        kwargs['type'] = self.__class__.__name__
        kwargs['id'] = InvalidId
        kwargs['parent'] = None
        return kwargs
    
    def add_job(self, job, callback=None, *args, **kw):
        th_pool = get_thread_pool()
        th_pool.add(job, callback, *args, **kw)

    # --------------------------------------------------------------------------
    # Environment DB related methods
    # --------------------------------------------------------------------------

    def set_environment_db(self, environment_db):
        """Sets the environment database.
        
        :param env_db:
            environment database name
        :type env_db:
            str
        """
        self.environment_manager.setEnvironmentDb(environment_db)

    # --------------------------------------------------------------------------
    # Python related methods
    # --------------------------------------------------------------------------

    def set_python_path(self, path):
        mod_man = ModuleManager()
        if self._path_id is not None:
            mod_man.remove_python_path(self._path_id)
        self._path_id = mod_man.add_python_path(path)

    # --------------------------------------------------------------------------
    # Macro path related methods
    # --------------------------------------------------------------------------

    def set_macro_path(self, macro_path):
        """Sets the macro path.
        
        :param macro_path:
            macro path
        :type macro_path:
            seq<str>
        """
        self.macro_manager.setMacroPath(macro_path)

    # --------------------------------------------------------------------------
    # Report related methods
    # --------------------------------------------------------------------------

    def set_log_report(self, filename=None, format=None):
        log = self.get_report_logger()
        
        # first check that the handler has not been initialized yet. If it has
        # we remove previous handlers. We only allow one timed rotating file
        # handler at a time
        to_remove = []
        for handler in log.handlers:
            if isinstance(handler, logging.handlers.TimedRotatingFileHandler):
                to_remove.append(handler)
        
        for handler in to_remove:
            handler.close()
            log.removeHandler(handler)
        
        if filename is None:
            return
        
        if format is None:
            format = Logger.DftLogMessageFormat
        formatter = logging.Formatter(format)
        
        self.info("Reports are being stored in %s", filename)
        klass = self.logReportKlass
        handler = klass(filename, **self.logReportParams)
        handler.setFormatter(formatter)
        log.addHandler(handler)
    
    def clear_log_report(self):
        self.set_log_report()
    
    def get_report_logger(self):
        return logging.getLogger("Sardana.Report")
    
    report_logger = property(get_report_logger)

    def report(self, msg, *args, **kwargs):
        """
        Record a log message in the sardana report (if enabled) with default
        level **INFO**. The msg is the message format string, and the args are
        the arguments which are merged into msg using the string formatting
        operator. (Note that this means that you can use keywords in the
        format string, together with a single dictionary argument.)
        
        *kwargs* are the same as :meth:`logging.Logger.debug` plus an optional
        level kwargs which has default value **INFO**
        
        Example::
        
            self.report("this is an official report!")
        
        :param msg: the message to be recorded
        :type msg: :obj:`str`
        :param args: list of arguments
        :param kwargs: list of keyword arguments"""
        level = kwargs.pop('level', logging.INFO)
        return self.report_logger.log(level, msg, *args, **kwargs)
    
    # --------------------------------------------------------------------------
    # Pool related methods
    # --------------------------------------------------------------------------
    
    def set_pool_names(self, pool_names):
        """Registers a new list of device pools in this manager
        
        :param pool_names: sequence of pool names
        :type pool_names: seq<str>"""
        for pool in self._pools.values():
            elements_attr = pool.getAttribute("Elements")
            elements_attr.removeListener(self.on_pool_elements_changed)
        
        for name in pool_names:
            self.debug("Creating pool %s", name)
            pool = Device(name)
            if pool is None:
                self.error('Could not create Pool object for %s' % name)
                continue
            self._pools[name] = pool
            elements_attr = pool.getAttribute("Elements")
            elements_attr.addListener(self.on_pool_elements_changed)
    
    def get_pool_names(self):
        """Returns the list of names of the pools this macro server is connected
        to.
        
        :return:
            the list of names of the pools this macro server is connected to
        :rtype:
            seq<str>"""
        return self._pools.keys()
    
    def get_pool(self, pool_name):
        """Returns the device pool object corresponding to the given device name
        or None if no match is found.
        
        :param pool_name: device pool name
        :type pool_name: str
        :return: Pool object or None if no match is found"""
        return self._pools.get(pool_name)
    
    def get_pools(self):
        """Returns the list of pools this macro server is connected to.
        
        :return: the list of pools this macro server is connected to
        :rtype: seq<Pool>"""
        return self._pools.values()
    
    def on_pool_elements_changed(self, evt_src, evt_type, evt_value):
        if evt_type not in CHANGE_EVT_TYPES:
            return
        self.fire_event(EventType("PoolElementsChanged"), evt_value)
    
    # --------------------------------------------------------------------------
    # Door related methods
    # --------------------------------------------------------------------------
    
    def create_element(self, **kwargs):
        type = kwargs['type']
        elem_type = ElementType[type]
        name = kwargs['name']
        
        kwargs['macro_server'] = self
        
        td = TYPE_MAP_OBJ[elem_type]
        klass = td.klass
        auto_full_name = td.auto_full_name
        
        full_name = kwargs.get("full_name", auto_full_name.format(**kwargs))
        
        self.check_element(name, full_name)
        
        id = kwargs.get('id')
        if id is None:
            kwargs['id'] = id = self.get_new_id()
        else:
            self.reserve_id(id)
        elem = klass(**kwargs)
        ret = self.add_element(elem)
        self.fire_event(EventType("ElementCreated"), elem)
        return ret
    
    def create_door(self, **kwargs):
        return self.create_element(type="Door", **kwargs)
        
    # --------------------------------------------------------------------------
    # General access to elements
    # --------------------------------------------------------------------------
    
    def get_elements_info(self):
        return self.get_remote_elements_info() + self.get_local_elements_info()
    
    def get_remote_elements_info(self):
        return [ elem.serialize()
            for pool in self.get_pools()
                for elem in pool.getElements() ]
    
    def get_local_elements_info(self):
        # fill macro library info
        ret = [ macrolib.serialize()
            for macrolib in self.get_macro_libs().values() ]
        # fill macro info
        ret += [ macro.serialize()
            for macro in self.get_macros().values() ]
        # fill parameter type info
        ret += [ paramtype.serialize()
            for paramtype in self.get_data_types().values() ]
        
        return ret
    
    # --------------------------------------------------------------------------
    # macro execution
    # --------------------------------------------------------------------------
    
    def set_max_parallel_macros(self, nb):
        assert nb > 0, "max parallel macros number must be > 0"
        th_pool = get_thread_pool()
        if th_pool.size + 5 < nb:
            th_pool.size = nb
        self._max_parallel_macros = nb
        
    def get_max_parallel_macros(self):
        return self._max_parallel_macros
    
    max_parallel_macros = property(get_max_parallel_macros,
        set_max_parallel_macros, doc="maximum number of macros which can "
        "execute at the same time")
    
    @property
    def macro_manager(self):
        return self._macro_manager
    
    @property
    def environment_manager(self):
        return self._environment_manager
    
    @property
    def type_manager(self):
        return self._type_manager
    
    # --------------------------------------------------------------------------
    # (Re)load code
    # --------------------------------------------------------------------------
    
    def reload_lib(self, lib_name):
        return self.macro_manager.reloadLib(lib_name)   
    
    def reload_macro_lib(self, lib_name):
        manager = self.macro_manager
        
        try:
            old_lib = manager.getMacroLib(lib_name)
        except UnknownMacroLibrary:
            old_lib = None
        
        new_elements, changed_elements, deleted_elements = [], [], []
        
        new_lib = manager.reloadMacroLib(lib_name)
        
        if old_lib is None:
            new_elements.extend(new_lib.get_macros())
            new_elements.append(new_lib)
        else:
            changed_elements.append(new_lib)
            new_names = set([ macro.name for macro in new_lib.get_macros() ])
            old_names = set([ macro.name for macro in old_lib.get_macros() ])
            changed_names = set.intersection(new_names, old_names)
            deleted_names = old_names.difference(new_names)
            new_names = new_names.difference(old_names)
            
            for new_name in new_names:
                new_elements.append(new_lib.get_macro(new_name))
            for changed_name in changed_names:
                changed_elements.append(new_lib.get_macro(changed_name))
            for deleted_name in deleted_names:
                deleted_elements.append(old_lib.get_macro(deleted_name))
        
        evt = { "new" : new_elements, "change" : changed_elements,
                "del" : deleted_elements }
        self.fire_event(EventType("ElementsChanged"), evt)
        return new_lib
        
    reload_macro_lib.__doc__ = MacroManager.reloadMacroLib.__doc__
    
    def reload_macro_libs(self, lib_names):
        for lib_name in lib_names:
            self.reload_macro_lib(lib_name)
    
    def reload_macro(self, macro_name):
        macro_info = self.macro_manager.getMacro(macro_name)
        lib_name = macro_info.module_name
        self.reload_macro_lib(lib_name)
    
    def reload_macros(self, macro_names):
        lib_names = set()
        for macro_name in macro_names:
            macro_info = self.macro_manager.getMacro(macro_name)
            lib_names.add(macro_info.module_name)
        self.reload_macro_libs(lib_names)
    
    def get_macro_lib(self, lib_name):
        return self.macro_manager.getMacroLib(lib_name)
    get_macro_lib.__doc__ = MacroManager.getMacroLib.__doc__
    
    def get_macro_libs(self, filter=None):
        return self.macro_manager.getMacroLibs(filter=filter)
    get_macro_libs.__doc__ = MacroManager.getMacroLibs.__doc__
    
    def get_macro_lib_names(self):
        return self.macro_manager.getMacroLibNames()
    get_macro_lib_names.__doc__ = MacroManager.getMacroLibNames.__doc__
    
    def get_macro(self, name):
        return self.macro_manager.getMacro(name)
    get_macro.__doc__ = MacroManager.getMacro.__doc__

    def get_macros(self, filter=None):
        return self.macro_manager.getMacros(filter=filter)
    get_macros.__doc__ = MacroManager.getMacros.__doc__
    
    def get_macro_names(self):
        return self.macro_manager.getMacroNames()
    get_macro_names.__doc__ = MacroManager.getMacroNames.__doc__
    
    def get_macro_classes(self):
        return self.macro_manager.getMacroClasses()
    get_macro_classes.__doc__ = MacroManager.getMacroClasses.__doc__
        
    def get_macro_functions(self):
        return self.macro_manager.getMacroFunctions()
    get_macro_functions.__doc__ = MacroManager.getMacroFunctions.__doc__
    
    def get_macro_libs_summary_info(self):
        libs = self.get_macro_libs()
        ret = []
        for module_name, macro_lib_info in libs.items():
            elem = "%s (%s)" % (macro_lib_info.name, macro_lib_info.file_path)
            ret.append(elem)
        return ret
    
    def get_macro_classes_summary_info(self):
        macros = self.get_macros()
        ret = []
        for macro_info in macros:
            elem = "%s (%s)" % (macro_info.name, macro_info.file_path)
            ret.append(elem)
        return ret
    
    def get_or_create_macro_lib(self, lib_name, macro_name=None):
        """Gets the exiting macro lib or creates a new macro lib file. If
        name is not None, a macro template code for the given macro name is 
        appended to the end of the file.
        
        :param lib_name:
            module name, python file name, or full file name (with path)
        :type lib_name: str
        :param macro_name:
            an optional macro name. If given a macro template code is appended
            to the end of the file (default is None meaning no macro code is
            added)
        :type macro_name: str
        
        :return:
            a sequence with three items: full_filename, code, line number is 0
            if no macro is created or n representing the first line of code for
            the given macro name.
        :rtype: seq<str, str, int>"""
        return self.macro_manager.getOrCreateMacroLib(lib_name,
                                                      macro_name=macro_name)
    get_or_create_macro_lib.__doc__ = MacroManager.getOrCreateMacroLib.__doc__
    
    def set_macro_lib(self, lib_name, code, auto_reload=True):
        module_name = self.macro_manager.setMacroLib(lib_name, code,
                                                     auto_reload=False)
        if auto_reload:
            self.reload_macro_lib(module_name)
        
    set_macro_lib.__doc__ = MacroManager.setMacroLib.__doc__
    
    # --------------------------------------------------------------------------
    # Data types
    # --------------------------------------------------------------------------
    
    def get_data_types(self):
        return self.type_manager.getTypes()
    get_data_types.__doc__ = TypeManager.getTypes.__doc__
    
    def get_data_type(self, type_name):
        return self.type_manager.getTypeObj(type_name)
    get_data_type.__doc__ = TypeManager.getTypeObj.__doc__
    
    def get_data_type_names(self):
        return self.type_manager.getTypeNames()
    get_data_type_names.__doc__ = TypeManager.getTypeNames.__doc__
    
    def get_data_type_names_with_asterisc(self):
        return self.type_manager.getTypeListStr()
    get_data_type_names_with_asterisc.__doc__ = TypeManager.getTypeListStr.__doc__
    
    # --------------------------------------------------------------------------
    # Doors
    # --------------------------------------------------------------------------
    
    def get_doors(self):
        return self.get_elements_by_type(ElementType.Door)
    
    def get_door_names(self):
        return [ door.full_name for door in self.get_doors() ]
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Environment access methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def get_env(self, key=None, door_name=None, macro_name=None):
        """Gets the environment matching the given parameters:
        
               - door_name and macro_name define the context where to look for
                 the environment. If both are None, the global environment is
                 used. If door name is None but macro name not, the given macro
                 environment is used and so on...
               - If key is None it returns the complete environment, otherwise
                 key must be a string containing the environment variable name.
        
        :param key:
            environment variable name [default: None, meaning all environment]
        :type key: str
        :param door_name:
            local context for a given door [default: None, meaning no door
            context is used]
        :type door_name: str
        :param macro_name:
            local context for a given macro [default: None, meaning no macro
            context is used]
        :type macro_name: str
        
        :return: a :obj:`dict` containing the environment
        :rtype: :obj:`dict`
        
        :raises: UnknownEnv"""
        return self.environment_manager.getEnv(key=key, macro_name=macro_name,
                                               door_name=door_name)
    
    def set_env(self, key, value):
        """Sets the environment key to the new value and stores it persistently.
        
        :param key: the key for the environment
        :param value: the value for the environment
        
        :return: a tuple with the key and value objects stored"""
        env_man = self.environment_manager
        if env_man.hasEnv(key):
            evt_type = "change"
        else:
            evt_type = "new"
        
        k,v = self.environment_manager.setEnv(key, value)
        
        evt = { evt_type : { k : v } }
        self.fire_event(EventType("EnvironmentChanged"), evt)
        return k,v
    
    def set_env_obj(self, data):
        """Sets the environment key to the new value and stores it persistently.
        
        :param key: the key for the environment
        :param value: the value for the environment
        
        :return: a tuple with the key and value objects stored"""
        env_man = self.environment_manager
        
        new, change = {}, {}
        for key, value in data.items():
            d = new
            if env_man.hasEnv(key):
                d = change
            d[key] = value
        
        ret = env_man.setEnvObj(data)
        
        evt = dict(new=new, change=change)
        self.fire_event(EventType("EnvironmentChanged"), evt)
        return ret
    
    def change_env(self, data):
        env_man = self.environment_manager
        new_change_env = data.get('new', {})
        new_change_env.update(data.get('change', {}))
        del_env = data.get('del', [])
        
        new, change = {}, {}
        for key, value in new_change_env.items():
            d = new
            if env_man.hasEnv(key):
                d = change
            d[key] = value
        
        del_keys = env_man.unsetEnv(del_env)
        env_man.setEnvObj(new_change_env)
        
        evt = dict(new=new, change=change)
        evt['del'] = del_keys
        self.fire_event(EventType("EnvironmentChanged"), evt)
    
    def unset_env(self, key):
        """Unsets the environment for the given key.
        
        :param key: the key for the environment to be unset"""
        ret = self.environment_manager.unsetEnv(key)
        evt = { 'del' : { key : None } }
        self.fire_event(EventType("EnvironmentChanged"), evt)
        return ret
        
    def has_env(self, key, macro_name=None, door_name=None):
        return self.environment_manager.hasEnv(key,
            macro_name=macro_name, door_name=door_name)
    
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # General object access methods
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    
    def get_object(self, name, type_class=All, subtype=All, pool=All):
        objs = self.find_objects(name, type_class, subtype, pool)
        if len(objs) == 0:
            return None
        if len(objs) > 1:
            raise AttributeError('More than one object named "%s" found' % name)
        return objs[0]

    def get_objects(self, names, type_class=All, subtype=All, pool=All):
        return self.find_objects(names, type_class=type_class, subtype=subtype,
                                 pool=pool)

    def find_objects(self, param, type_class=All, subtype=All, pool=All):
        if is_pure_str(param):
            param = param,
        
        if type_class == MacroServer.All:
            type_name_list = self.get_data_type_names()
        else:
            if is_pure_str(type_class):
                type_name_list = type_class,
            else:
                type_name_list = type_class
        obj_set = set()
        param = [ '^%s$' % x for x in param ]
        re_objs = map(re.compile, param, len(param)*(re.IGNORECASE,))
        re_subtype = re.compile(subtype, re.IGNORECASE)
        for type_name in type_name_list:
            type_class_name = type_name
            if type_class_name.endswith('*'):
               type_class_name = type_class_name[:-1]
            type_inst = self.get_data_type(type_class_name)
            if not type_inst.hasCapability(ParamType.ItemList):
                continue
            if self.is_macroserver_interface(type_class_name):
                for name, obj in type_inst.getObjDict(pool=pool).items():
                    for re_obj in re_objs:
                        if re_obj.match(name) is not None:
                            obj_type = ElementType[obj.get_type()]
                            if subtype is MacroServer.All or re_subtype.match(obj_type):
                                obj_set.add(obj)
            else:
                for name, obj in type_inst.getObjDict(pool=pool).items():
                    for re_obj in re_objs:
                        if re_obj.match(name) is not None:
                            obj_type = obj.getType()
                            if (subtype is MacroServer.All or \
                                re_subtype.match(obj.getType())) and \
                               obj_type != "MotorGroup":
                                obj_set.add(obj)
        return list(obj_set)
    
    def get_motion(self, elems, motion_source=None, read_only=False, cache=True,
                   decoupled=False):
        if motion_source is None:
            motion_source = self.get_pools()
        
        motion_klass = Motion
        if decoupled: # and len(elems)>1:
            motion_klass = MotionGroup
        return motion_klass(elems, motion_source)
    
    _LOCAL_INTERFACES = {
        Interface.MacroLibrary : get_macro_libs,
        Interface.MacroCode : get_macros,
        Interface.MacroClass: get_macro_classes,
        Interface.MacroFunction: get_macro_functions, 
    }
    
    def is_macroserver_interface(self, interface):
        if is_pure_str(interface):
            interface = Interface[interface]
        return interface in self._LOCAL_INTERFACES
    
    def get_elements_with_interface(self, interface):
        ret=CaselessDict()
        if is_pure_str(interface):
            interface_str = interface
            interface = Interface[interface_str]
        else:
            interface_str = Interface[interface]

        if self.is_macroserver_interface(interface):
            ret.update(self._LOCAL_INTERFACES.get(interface)(self))
        else:
            for pool in self.get_pools():
                ret.update(pool.getElementsWithInterface(interface_str))
        return ret
    
    def get_element_with_interface(self, name, interface):
        for pool in self.get_pools():
            element = pool.getElementWithInterface(name, interface)
            if element is not None:
                return element
    
    def get_controllers(self):
        return self.get_elements_with_interface("Controller")
    
    def get_moveables(self):
        return self.get_elements_with_interface("Moveable")
    
    def get_motors(self):
        return self.get_elements_with_interface("Motor")
    
    def get_pseudo_motors(self):
        return self.get_elements_with_interface("PseudoMotor")
    
    def get_io_registers(self):
        return self.get_elements_with_interface("IORegister")
    
    def get_measurement_groups(self):
        return self.get_elements_with_interface("MeasurementGroup")
    
    def get_exp_channels(self):
        return self.get_elements_with_interface("ExpChannel")
        
    def get_counter_timers(self):
        return self.get_elements_with_interface("CTExpChannel")
    
    def get_0d_exp_channels(self):
        return self.get_elements_with_interface("ZeroDExpChannel")
    
    def get_1d_exp_channels(self):
        return self.get_elements_with_interface("OneDExpChannel")
    
    def get_2d_exp_channels(self):
        return self.get_elements_with_interface("TwoDExpChannel")
    
    def get_pseudo_counters(self):
        return self.get_elements_with_interface("PseudoCounter")
    
    def get_instruments(self):
        return self.get_elements_with_interface("Instrument")
    
    def get_controller(self, name):
        return self.get_element_with_interface(name, "Controller")
    
    def get_moveable(self, name):
        return self.get_element_with_interface(name, "Moveable")
    
    def get_motor(self, name):
        return self.get_element_with_interface(name, "Motor")
    
    def get_pseudo_motor(self, name):
        return self.get_element_with_interface(name, "PseudoMotor")
    
    def get_io_register(self, name):
        return self.get_element_with_interface(name, "IORegister")
    
    def get_measurement_group(self, name):
        return self.get_element_with_interface(name, "MeasurementGroup")
    
    def get_exp_channel(self, name):
        return self.get_element_with_interface(name, "ExpChannel")
        
    def get_counter_timer(self, name):
        return self.get_element_with_interface(name, "CTExpChannel")
    
    def get_0d_exp_channel(self, name):
        return self.get_element_with_interface(name, "ZeroDExpChannel")
    
    def get_1d_exp_channel(self, name):
        return self.get_element_with_interface(name, "OneDExpChannel")
    
    def get_2d_exp_channel(self, name):
        return self.get_element_with_interface(name, "TwoDExpChannel")
    
    def get_pseudo_counter(self, name):
        return self.get_element_with_interface(name, "PseudoCounter")
    
    def get_instrument(self, name):
        return self.get_element_with_interface(name, "Instrument")



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import math

import PyTango

from taurus.core.util import CaselessDict

from sardana import State, DataAccess
from sardana.pool.controller import MotorController, CounterTimerController, \
    ZeroDController, \
    Type, Access, Description, DefaultValue

TangoAttribute = "TangoAttribute"
Formula = "Formula"

class ReadTangoAttributes(object):
    """ Generic class that has as many devices as the user wants.
    Each device has a tango attribute and a formula and the 'hardware' tango calls
    are optimized in the sense that only one call per tango device is issued.
    """
    axis_attributes = {
        TangoAttribute : { Type : str, Access : DataAccess.ReadWrite,
                           Description : 'Attribute to read (e.g. a/b/c/attr)' },
        Formula        : { Type : str, Access : DataAccess.ReadWrite,
                           DefaultValue : "VALUE",
                           Description  : 'The Formula to get the desired value.\n'
                                          'e.g. "math.sqrt(VALUE)"' },
    }
    
    def __init__(self):
        #: dict<int(axis), str(reason for being in pending)>
        self._pending = {}
        
        #: dict<str(dev name), tuple<DeviceProxy, list<str(attributes name)>>>
        self._devices = CaselessDict()
        
        #: dict<int(axis), seq<str<tango full attribute name>, str<attr name>, DeviceProxy>>
        self._axis_tango_attributes = {}
        
        #: dict<int(axis), str<formula>>
        self._axis_formulas = {}
    
    def add_device(self, axis):
        self._pending[axis] = "No tango attribute associated to this device yet"
        self._axis_formulas[axis] = self.axis_attribute[Formula][DefaultValue]
    
    def delete_device(self, axis):
        if axis in self._pending:
            del self._pending[axis]
        else:
            del self._axis_tango_attributes[axis]
            del self._axis_formulas[axis]
    
    def state_one(self, axis):
        pending_info = self._pending.get(axis)
        if pending_info is not None:
            return State.Fault, pending_info
        return State.On, 'Always ON, just reading tango attribute'
    
    def pre_read_all(self):
        self._devices_read = {}
    
    def pre_read_one(self, axis):
        attr_name, dev = self._axis_tango_attributes[axis][1:]
        dev_attrs = self._devices_read.get(dev)
        if dev_attrs is None:
            self._
            self._devices_read[dev] = dev_attrs = []
        dev_attrs.append(attr_name)
    
    def read_all(self):
        pass
    
    def read_one(self, axis):
        pass
    
    def get_extra_attribute_par(self, axis, name):
        if name == TangoAttribute:
            return self._axis_tango_attributes[axis][0]
        elif name == Formula:
            return self._axis_formulas[axis]
    
    def set_extra_attribute_par(self, axis, name, value):
        if name == TangoAttribute:
            value = value.lower()
            self._axis_tango_attributes[axis] = data = value, None, None
            try:
                dev_name, attr_name = value.rsplit("/",1)
                data[1] = attr_name
            except:
                self._pending[axis] = "invalid device name " + value
                raise Exception(self._pending[axis])
            dev_info = self._devices.get(dev_name)
            if dev_info is None:
                try:
                    proxy = PyTango.DeviceProxy(dev_name)
                except PyTango.DevFailed, df:
                    if len(df):
                        self._pending[axis] = df[0].reason + ": " + df[0].desc
                    else:
                        self._pending[axis] = "Unknwon PyTango Error: " + str(df)
                    raise
                self._devices[dev_name] = dev_info = proxy, []
            data[2] = dev_info[0]
            dev_info[1].append(attr_name)
            
        elif name == Formula:
            self._axis_formulas[axis] = value

class TangoCounterTimerController(ReadTangoAttributes, CounterTimerController):
    """This controller offers as many channels as the user wants.
    Each channel has two _MUST_HAVE_ extra attributes:
    +) TangoAttribute - Tango attribute to retrieve the value of the counter
    +) Formula - Formula to evaluate using 'VALUE' as the tango attribute value
    As examples you could have:
    ch1.TangoExtraAttribute = 'my/tango/device/attribute1'
    ch1.Formula = '-1 * VALUE'
    ch2.TangoExtraAttribute = 'my/tango/device/attribute2'
    ch2.Formula = 'math.sqrt(VALUE)'
    ch3.TangoExtraAttribute = 'my_other/tango/device/attribute1'
    ch3.Formula = 'math.cos(VALUE)'
    """

    gender = ""
    model  = ""
    organization = "Sardana team"

    MaxDevice = 1024

    def __init__(self, inst, props, *args, **kwargs):
        ReadTangoAttributes.__init__(self)
        CounterTimerController.__init__(self, inst, props, *args, **kwargs)
    
    def AddDevice(self, axis):
        self.add_device(axis)

    def DeleteDevice(self, axis):
        self.delete_device(axis)

    def StateOne(self, axis):
        return self.state_one(axis)

    def PreReadAll(self):
        self.pre_read_all()
        
    def PreReadOne(self, axis):
        self.pre_read_one(axis)

    def ReadAll(self):
        self.read_all()

    def ReadOne(self, axis):
        return self.read_one(axis)

    def GetExtraAttributePar(self, axis, name):
        return self.get_extra_attribute_par(axis, name)

    def SetExtraAttributePar(self,axis, name, value):
        self.set_extra_attribute_par(axis, name, value)
        
    def SendToCtrl(self,in_data):
        return ""
    
    def AbortOne(self, axis):
        pass
        
    def PreStartAllCT(self):
        pass
    
    def StartOneCT(self, axis):
        pass
    
    def StartAllCT(self):
        pass
    
    def LoadOne(self, axis, value):
        pass




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the class for a
motion"""

__all__ = ["MotionState", "MotionMap", "PoolMotion", "PoolMotionItem"]

__docformat__ = 'restructuredtext'

import time

from taurus.core.util import Enumeration, DebugIt

from sardana import State
from poolaction import ActionContext, PoolActionItem, PoolAction

#: enumeration representing possible motion states
MotionState = Enumeration("MotionSate", ( \
    "Stopped",
#    "StoppedOnError",
#    "StoppedOnAbort",
    "Moving",
    "MovingBacklash",
    "MovingInstability",
    "Invalid"))

MS = MotionState
MovingStates = MS.Moving, MS.MovingBacklash, MS.MovingInstability
StoppedStates = MS.Stopped, #MS.StoppedOnError, MS.StoppedOnAbort

#MotionAction = Enumeration("MotionAction", ( \
#    "StartMotion",
#    "Finish",
#    "Abort",
#    "NoAction",
#    "Invalid") )

#MA = MotionAction

MotionMap = {
    #MS.Stopped          : State.On,
    MS.Moving            : State.Moving,
    MS.MovingBacklash    : State.Moving,
    MS.MovingInstability : State.Moving,
    MS.Invalid           : State.Invalid,
}


class PoolMotionItem(PoolActionItem):
    """An item involved in the motion. Maps directly to a motor object"""

    def __init__(self, moveable, position, dial_position, do_backlash,
                 backlash, instability_time=None):
        PoolActionItem.__init__(self, moveable)
        self.position = position
        self.interrupted = False
        self.dial_position = dial_position
        self.do_backlash = do_backlash
        self.backlash = backlash
        self.instability_time = instability_time
        self.old_motion_state = MS.Invalid
        self.motion_state = MS.Stopped
        self.start_time = None
        self.stop_time = None
        self.stop_final_time = None
        self.old_state_info = State.Invalid, "Uninitialized", \
            (False, False, False)
        self.state_info = State.On, "Uninitialized", (False, False, False)

    def has_instability_time(self):
        return self.instability_time is not None

    def in_motion(self):
        return self.motion_state in MovingStates

    def get_moveable(self):
        return self.element

    moveable = property(fget=get_moveable)

    def get_state_info(self):
        si = self.state_info
        return MotionMap.get(self.motion_state, si[0]), si[1], si[2]

    def start(self, new_state):
        self.old_state_info = self.state_info
        self.state_info = new_state, self.state_info[1], self.state_info[2]

    def stopped(self, timestamp):
        self.stop_time = timestamp
        if self.instability_time is None:
            self.stop_final_time = timestamp
            new_ms = MS.Stopped
        else:
            new_ms = MS.MovingInstability
        return new_ms

    def handle_instability(self, timestamp):
        new_ms = MS.MovingInstability
        dt = timestamp - self.stop_time
        if dt >= self.instability_time:
            self.stop_final_time = timestamp
            new_ms = MS.Stopped
        return new_ms

    def on_state_switch(self, state_info, timestamp=None):
        if timestamp is None:
            timestamp = time.time()
        self.old_state_info = self.state_info
        self.state_info = state_info
        old_state = self.old_state_info[0]
        state = state_info[0]
        new_ms = ms = self.motion_state
        moveable = self.moveable
        self.interrupted = moveable.was_interrupted()

        if self.interrupted:
            if ms == MS.MovingInstability:
                new_ms = self.handle_instability(timestamp)
            elif state == State.Moving:
                new_ms = MS.Moving
            elif old_state == State.Moving:
                new_ms = self.stopped(timestamp)
        elif ms == MS.Stopped:
            if state == State.Moving:
                self.start_time = timestamp
                new_ms = MS.Moving
        elif ms == MS.Moving:
            if state != State.Moving:
                if self.do_backlash and state == State.On:
                    new_ms = MS.MovingBacklash
                else:
                    new_ms = self.stopped(timestamp)
        elif ms == MS.MovingBacklash:
            if state != State.Moving:
                new_ms = self.stopped(timestamp)
        elif ms == MS.MovingInstability:
            new_ms = self.handle_instability(timestamp)
        self.old_motion_state, self.motion_state = ms, new_ms
        return ms, new_ms

_NON_ERROR_STATES = State.On, State.Moving, State.Running

class PoolMotion(PoolAction):
    """This class manages motion actions"""

    def __init__(self, main_element, name="GlobalMotion"):
        PoolAction.__init__(self, main_element, name)
        self._motion_info = None
        self._motion_sleep_time = None
        self._nb_states_per_position = None

    def _recover_start_error(self, ctrl, meth_name, read_state=False):
        self.error("%s throws exception on %s. Stopping...", ctrl, meth_name)
        self.debug("Details:", exc_info=1)

        self.emergency_break()

        if read_state:
            states = {}
            self.read_state_info(ret=states)
            for moveable, state_info in states.items():
                state_info = moveable._from_ctrl_state_info(state_info)
                moveable._set_state_info(state_info)

    def pre_start_all(self, pool_ctrls):
        # PreStartAll on all controllers
        for pool_ctrl in pool_ctrls:
            try:
                pool_ctrl.ctrl.PreStartAll()
            except:
                self._recover_start_error(pool_ctrl, "PreStartAll")
                raise

    def pre_start_one(self, moveables, items):
        # PreStartOne on all elements
        for moveable in moveables:
            pool_ctrl = moveable.controller
            ctrl, axis = pool_ctrl.ctrl, moveable.axis
            dial = items[moveable][1]
            ret = ctrl.PreStartOne(axis, dial)
            if not ret:
                try:
                    msg = "%s.PreStartOne(%s(%d), %f) returns False" \
                           % (pool_ctrl.name, moveable.name, axis, dial)
                    raise Exception(msg)
                except:
                    self._recover_start_error(pool_ctrl, "PreStartOne")
                    raise

    def start_one(self, moveables, motion_info):
        # StartOne on all elements
        for moveable in moveables:
            pool_ctrl = moveable.controller
            ctrl = pool_ctrl.ctrl
            axis = moveable.axis
            dial_position = motion_info[moveable].dial_position
            try:
                ctrl.StartOne(axis, dial_position)
            except:
                self._recover_start_error(pool_ctrl, "StartOne")
                raise

    def start_all(self, pool_ctrls, moveables, motion_info):
        # Change the state to Moving
        for moveable in moveables:
            moveable_info = motion_info[moveable]
            moveable.set_state(State.Moving, propagate=2)
            state_info = moveable.inspect_state(), \
                moveable.inspect_status(), \
                moveable.inspect_limit_switches()
            moveable_info.on_state_switch(state_info)

        # StartAll on all controllers
        for pool_ctrl in pool_ctrls:
            try:
                pool_ctrl.ctrl.StartAll()
            except:
                self._recover_start_error(pool_ctrl, "StartOne",
                                          read_state=True)
                raise

    def start_action(self, *args, **kwargs):
        """kwargs['items'] is a dict<moveable, (pos, dial, do_backlash, backlash)
        """
        items = kwargs.pop("items")

        pool = self.pool

        # prepare data structures
        self._aborted = False
        self._stopped = False
        self._motion_sleep_time = kwargs.pop("motion_sleep_time",
                                             pool.motion_loop_sleep_time)
        self._nb_states_per_position = \
            kwargs.pop("nb_states_per_position",
                       pool.motion_loop_states_per_position)

        self._motion_info = motion_info = {}
        for moveable, motion_data in items.items():
            it = moveable.instability_time
            motion_info[moveable] = PoolMotionItem(moveable, *motion_data,
                                                   instability_time=it)

        pool_ctrls = self.get_pool_controller_list()
        moveables = self.get_elements()

        with ActionContext(self):
            self.pre_start_all(pool_ctrls)
            self.pre_start_one(moveables, items)
            self.start_one(moveables, motion_info)
            self.start_all(pool_ctrls, moveables, motion_info)

    def backlash_item(self, motion_item):
        moveable = motion_item.moveable
        controller = moveable.controller
        axis = moveable.axis
        position = motion_item.backlash
        try:
            controller.move({axis: position})
        except:
            self.warning("could not start backlash on %s", moveable.name,
                         exc_info=1)

    @DebugIt()
    def action_loop(self):
        i = 0
        states, positions = {}, {}
        for k in self.get_elements():
            states[k] = None
            positions[k] = None

        nap = self._motion_sleep_time
        nb_states_per_pos = self._nb_states_per_position
        motion_info = self._motion_info
        emergency_stop = set()

        # read positions to send a first event when starting to move
        #with ActionContext(self) as context:
        #    positions = self.raw_read_dial_position()
        #    position_error_occured = self._position_error_occured(positions)
        #    if position_error_occured:
        #        self.error("Pre-loop read position error")
        #        self.emergency_break()
        #        for moveable, position_info in positions.items():
        #            if position_info[1] is not None:
        #                self.error("%s error reading position",moveable.name)
        #                self.debug("Details:", exc_info=position_info[1])
        #
        #        # send position
        #        for moveable, position_info in positions.items():
        #            moveable.put_dial_position(position_info, propagate=2)
        #            position = moveable.get_position(propagate=0)
        #            if position.in_error():
        #                emergency_stop.add(moveable)
        #
        #        # send state
        #        self.read_state_info(ret=states)
        #        for moveable, state_info in states.items():
        #            state_info = moveable._from_ctrl_state_info(state_info)
        #            state_info = list(state_info)
        #            state_info[0] = State.Fault
        #            moveable.set_state_info(state_info, propagate=2)
        #        return
        #    else:
        #        for moveable, position_info in positions.items():
        #            moveable.put_dial_position(position_info)

        while True:
            self.read_state_info(ret=states)
            state_error_occured = self._state_error_occured(states)
            timestamp = time.time()
            in_motion = False
            for moveable, state_info in states.items():
                motion_item = motion_info[moveable]
                state_info = moveable._from_ctrl_state_info(state_info)

                state, status, limit_switches = state_info
                old_motion_state, motion_state = \
                    motion_item.on_state_switch(state_info,
                                                timestamp=timestamp)
                real_state_info = motion_item.get_state_info()
                moving = motion_item.in_motion()

                # Something wrong happened: Stop all motors and report
                if state_error_occured:
                    in_motion = False
                    emergency_stop.add(moveable)

                start_backlash = motion_state == MS.MovingBacklash and \
                    old_motion_state != MS.MovingBacklash
                start_instability = motion_state == MS.MovingInstability and \
                    old_motion_state != MS.MovingInstability
                stopped_now = not moving and old_motion_state in MovingStates

                # make sure the state is placed in the motor
                moveable.put_state_info(real_state_info)

                if emergency_stop:
                    continue

                # if motor stopped 'well' and there is a backlash to do...
                if start_backlash:
                    moveable.debug("Starting backlash")
                    # make sure the last position after the first motion is
                    # sent before starting the backlash motion
                    moveable.get_position(cache=False, propagate=2)
                    self.backlash_item(motion_item)
                    moving = motion_item.in_motion()
                elif start_instability:
                    moveable.debug("Starting to wait for instability")
                    # make sure the last position after the first motion is
                    # sent before starting the backlash motion
                    moveable.get_position(cache=False, propagate=2)
                elif stopped_now:
                    moveable.debug("Stopped")

                    # try to read a last position to force an event
                    moveable.get_position(cache=False, propagate=2)

                    ## free the motor from the OperationContext so we can send
                    ## a state event. Otherwise we may be asked to move the
                    ## motor again which would result in an exception saying
                    ## that the motor is already involved in an operation
                    ## ... but before protect the motor so that the monitor
                    ## doesn't come in between the two instructions below and
                    ## send a state event on it's own
                    #with moveable:
                    #    moveable.clear_operation()
                    moveable.set_state_info(real_state_info, propagate=2)

                # Then update the state
                if not stopped_now:
                    moveable.set_state_info(real_state_info, propagate=1)

                if moving:
                    in_motion = True

            if not in_motion:
                if state_error_occured:
                    self.error("Loop read state error")
                    self.emergency_break()

                    # send positions
                    self.read_dial_position(ret=positions)
                    position_error_occured = self._position_error_occured(positions)
                    if position_error_occured:
                        self.error("Loop final read position error. "
                                   "Retrying...")
                        self.read_dial_position(ret=positions)
                        position_error_occured = self._position_error_occured(positions)
                        if position_error_occured:
                            self.error("Loop final read position error 2. "
                                       "Cannot send final position event!!!")

                    for moveable, position_info in positions.items():
                        moveable.put_dial_position(position_info, propagate=2)

                    # send state
                    for moveable in states:
                        if moveable not in emergency_stop:
                            moveable.get_state(cache=False, propagate=2)
                        else:
                            motion_item = motion_info[moveable]
                            real_state_info = motion_item.get_state_info()
                            moveable.set_state_info(real_state_info,
                                                    propagate=2)
                break

            # read position every n times
            if not i % nb_states_per_pos:
                self.read_dial_position(ret=positions)
                # send position
                for moveable, position_value in positions.items():
                    if position_value.error:
                        self.error("Loop read position error for %s" % moveable.name)
                    moveable.put_dial_position(position_value)
            i += 1
            time.sleep(nap)

    def _state_error_occured(self, d):
        for _, (state_info, exc_info) in d.items():
            state = state_info[0]
            if exc_info is not None or state not in _NON_ERROR_STATES:
                return True
        return False

    def _position_error_occured(self, positions):
        for _, value in positions.items():
            if value.error:
                return True

    def _recover_moving_error(self, location, emergency_stop):
        emergency_names = [moveable.name for moveable in emergency_stop]
        self.error("%s: error on %s", location, emergency_names)

        # stop everything
        self.emergency_break()

    def _recover_position_moving_error(self, location, emergency_stop,
                                       positions):
        self._recover_moving_error(location, emergency_stop)

        # send state
        states = {}
        self.read_state_info(ret=states)
        motion_info = self._motion_info
        for moveable in states:
            if moveable not in emergency_stop:
                moveable.get_state(cache=False, propagate=2)
            else:
                motion_item = motion_info[moveable]
                real_state_info = motion_item.get_state_info()
                moveable.set_state_info(real_state_info, propagate=2)

    def _recover_state_moving_error(self, location, emergency_stop, states):
        self._recover_moving_error(location, emergency_stop)

        # send positions
        positions = {}
        self.read_dial_position(ret=positions)
        for moveable, position_info in positions.items():
            moveable.put_dial_position(position_info, propagate=2)

        motion_info = self._motion_info
        # send state
        for moveable in states:
            if moveable not in emergency_stop:
                moveable.get_state(cache=False, propagate=2)
            else:
                motion_item = motion_info[moveable]
                real_state_info = motion_item.get_state_info()
                moveable.set_state_info(real_state_info, propagate=2)

    def read_dial_position(self, ret=None, serial=False):
        return self.read_value(ret=ret, serial=serial)

    def raw_read_dial_position(self, ret=None, serial=False):
        return self.raw_read_value(ret=ret, serial=serial)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Sardana
## 
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package provides the sardana library"""

__all__ = ['make_color_table', 'TermColors', 'HTMLColors', 'TermTangoColors',
           'TermTauSWDevStateColors', 'HTMLTangoColors']

__docformat__ = 'restructuredtext'

def make_color_table(in_class,use_name=False):
    """Build a set of color attributes in a class.

    Helper function for building the *TermColors classes."""
    
    color_templates = (
        ("Black"       , "0;30"),
        ("Red"         , "0;31"),
        ("Green"       , "0;32"),
        ("Brown"       , "0;33"),
        ("Blue"        , "0;34"),
        ("Purple"      , "0;35"),
        ("Cyan"        , "0;36"),
        ("LightGray"   , "0;37"),
        ("DarkGray"    , "1;30"),
        ("LightRed"    , "1;31"),
        ("LightGreen"  , "1;32"),
        ("Yellow"      , "1;33"),
        ("LightBlue"   , "1;34"),
        ("LightPurple" , "1;35"),
        ("LightCyan"   , "1;36"),
        ("White"       , "1;37"),  )

    for name,value in color_templates:
        if use_name:
            setattr(in_class,name,in_class._base % name)
        else:
            setattr(in_class,name,in_class._base % value)

class TermColors:
    """Color escape sequences.

    This class defines the escape sequences for all the standard (ANSI?) 
    colors in terminals. Also defines a NoColor escape which is just the null
    string, suitable for defining 'dummy' color schemes in terminals which get
    confused by color escapes.

    This class should be used as a mixin for building color schemes.
    
    Basicaly this class is just a copy of IPython.ColorANSI.TermColors class"""
    
    NoColor = ''  # for color schemes in color-less terminals.
    Normal = '\033[0m'   # Reset normal coloring
    _base  = '\033[%sm'  # Template for all other colors

class HTMLColors:
    
    NoColor = ''
    Normal = '</font>'
    _base = '<font color=%s>'

# Build the actual color table as a set of class attributes:
make_color_table(TermColors)
make_color_table(HTMLColors,True)

import PyTango

TermTangoColors = { PyTango.DevState.ON      : TermColors.Green,
                    PyTango.DevState.ALARM   : TermColors.Brown,
                    PyTango.DevState.FAULT   : TermColors.Red,
                    PyTango.DevState.UNKNOWN : TermColors.LightGray,
                    None                     : TermColors.DarkGray }

from taurus.core import TaurusSWDevState

TermTauSWDevStateColors = {
    TaurusSWDevState.Uninitialized       : TermColors.LightGray,
    TaurusSWDevState.Running             : TermColors.Green, 
    TaurusSWDevState.Shutdown            : TermColors.DarkGray, 
    TaurusSWDevState.Crash               : TermColors.Red,
    TaurusSWDevState.EventSystemShutdown : TermColors.Brown }
    
HTMLTangoColors = { PyTango.DevState.ON      : HTMLColors.Green,
                    PyTango.DevState.ALARM   : HTMLColors.Brown,
                    PyTango.DevState.FAULT   : HTMLColors.Red,
                    PyTango.DevState.UNKNOWN : HTMLColors.LightGray,
                    None                     : HTMLColors.DarkGray }


#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Generic Sardana Tango device module"""

from __future__ import with_statement

__all__ = ["SardanaDevice", "SardanaDeviceClass"]

__docformat__ = 'restructuredtext'

import time
import threading

import PyTango.constants
from PyTango import Device_4Impl, DeviceClass, Util, DevState, \
    AttrQuality, TimeVal, ArgType, ApiUtil, DevFailed, WAttribute

from taurus.core.util.threadpool import ThreadPool
from taurus.core.util.log import Logger

from util import to_tango_state, NO_DB_MAP


__thread_pool_lock = threading.Lock()
__thread_pool = None

def get_thread_pool():
    """Returns the global pool of threads for Sardana

    :return: the global pool of threads object
    :rtype: taurus.core.util.ThreadPool"""

    global __thread_pool

    if __thread_pool:
        return __thread_pool

    global __thread_pool_lock
    with __thread_pool_lock:
        if __thread_pool is None:
            __thread_pool = ThreadPool(name="EventTH", Psize=1, Qsize=1000)
        return __thread_pool


class SardanaDevice(Device_4Impl, Logger):
    """SardanaDevice represents the base class for all Sardana 
    :class:`PyTango.DeviceImpl` classes"""
    
    def __init__(self, dclass, name):
        """Constructor"""
        self.in_constructor = True
        try:
            Device_4Impl.__init__(self, dclass, name)
            self.init(name)
            Logger.__init__(self, name)
    
            self._state = DevState.INIT
            self._status = 'Waiting to be initialized...'
    
            # access to some tango API (like MultiAttribute and Attribute) is
            # still not thread safe so we have this lock to protect
            # Wa can't always use methods which use internally the
            # C++ AutoTangoMonitor because it blocks the entire tango device.
            self.tango_lock = threading.RLock()
    
            self._event_thread_pool = get_thread_pool()
            self.init_device()
        finally:
            self.in_constructor = False

    def init(self, name):
        """initialize the device once in the object lifetime. Override when
        necessary but **always** call the method from your super class
        
        :param str name: device name"""
        
        db = self.get_database()
        if db is None:
            self._alias = self._get_nodb_device_info()[0]
        else:
            try:
                self._alias = db.get_alias(name)
                if self._alias.lower() == 'nada':
                    self._alias = None
            except:
                self._alias = None

    def get_alias(self):
        """Returns this device alias name

        :return: this device alias
        :rtype: str"""
        return self._alias

    alias = property(get_alias, doc="the device alias name")

    def get_full_name(self):
        """Returns the device full name in format
        dbname:dbport/<domain>/<family>/<member>
        
        :return: this device full name
        :rtype: str"""
        db = self.get_database()
        if db.get_from_env_var():
            db_name = ApiUtil.get_env_var("TANGO_HOST")
        else:
            if db.is_dbase_used():
                db_name = db.get_db_host() + ":" + db.get_db_port()
            else:
                db_name = db.get_file_name()
        return db_name + "/" + self.get_name()

    def init_device(self):
        """Initialize the device. Called during startup after :meth:`init` and
        every time the tango ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        self.set_state(self._state)
        db = self.get_database()
        if db is None:
            self.init_device_nodb()
        else:
            self.get_device_properties(self.get_device_class())

        detect_evts = "state", "status"
        non_detect_evts = ()
        self.set_change_events(detect_evts, non_detect_evts)

    def _get_nodb_device_info(self):
        """Internal method. Returns the device info when tango database is not
        being used (example: in demos)"""
        name = self.get_name()
        tango_class = self.get_device_class().get_name()
        devices = NO_DB_MAP.get(tango_class, ())
        for dev_info in devices:
            if dev_info[1] == name:
                return dev_info

    def init_device_nodb(self):
        """Internal method. Initialize the device when tango database is not
        being used (example: in demos)"""
        _, _, props = self._get_nodb_device_info()
        for prop_name, prop_value in props.items():
            setattr(self, prop_name, prop_value)

    def delete_device(self):
        """Clean the device. Called during shutdown and every time the tango
        ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        pass

    def set_change_events(self, evts_checked, evts_not_checked):
        """Helper method to set change events on attributes
        
        :param evts_checked:
            list of attribute names to activate change events programatically
            with tango filter active
        :type evts_checked: seq<:obj:`str`\>
        :param evts_not_checked:
            list of attribute names to activate change events programatically
            with tango filter inactive. Use this with care! Attributes
            configured with no change event filter may potentially generated a
            lot of events!
        :type evts_not_checked: seq<:obj:`str`\>"""     
        for evt in evts_checked:
            self.set_change_event(evt, True, True)
        for evt in evts_not_checked:
            self.set_change_event(evt, True, False)

    def initialize_dynamic_attributes(self):
        """Initialize dynamic attributes. Default implementation does nothing.
        Override when necessary."""
        pass

    def get_event_thread_pool(self):
        """Return the :class:`~taurus.core.util.ThreadPool` used by sardana to
        send tango events.
        
        :return: the sardana :class:`~taurus.core.util.ThreadPool`
        :rtype: :class:`~taurus.core.util.ThreadPool`"""
        return self._event_thread_pool
    
    def get_attribute_by_name(self, attr_name):
        """Gets the attribute for the given name.
        
        :param attr_name: attribute name
        :type attr_name: str
        :return: the attribute object
        :rtype: :class:`~PyTango.Attribute`"""
        return self.get_device_attr().get_attr_by_name(attr_name)
    
    def get_wattribute_by_name(self, attr_name):
        """Gets the writable attribute for the given name.
        
        :param attr_name: attribute name
        :type attr_name: str
        :return: the attribute object
        :rtype: :class:`~PyTango.WAttribute`"""
        return self.get_device_attr().get_w_attr_by_name(attr_name)
    
    def get_database(self):
        """Helper method to return a reference to the current tango database
        
        :return: the Tango database
        :rtype: :class:`~PyTango.Database`"""
        return Util.instance().get_database()
    
    def set_write_attribute(self, attr, w_value):
        try:
            attr.set_write_value(w_value)
        except DevFailed as df:
            df0 = df[0]
            reason = df0.reason
            # if outside limit prefix the description with the device name
            if reason == PyTango.constants.API_WAttrOutsideLimit:
                desc = self.alias + ": " + df0.desc
                _df = DevFailed(*df[1:])
                PyTango.Except.re_throw_exception(_df, df0.reason, desc, df0.origin)
            raise df
    
    def set_attribute(self, attr, value=None, w_value=None, timestamp=None,
                      quality=None, error=None, priority=1, synch=True):
        """Sets the given attribute value. If timestamp is not given, *now* is
        used as timestamp. If quality is not given VALID is assigned. If error
        is given an error event is sent (with no value and quality INVALID).
        If priority is > 1, the event filter is temporarily disabled so the event
        is sent for sure. If synch is set to True, wait for fire event to finish
        
        :param attr: 
            the tango attribute
        :type attr: :class:`PyTango.Attribute`
        :param value:
            the value to be set (not mandatory if setting an error)
            [default: None]
        :type value: object
        :param w_value:
            the write value to be set (not mandatory)
            [default: None, meaning maintain current write value]
        :type value: object
        :param timestamp:
            the timestamp associated with the operation [default: None, meaning
            use *now* as timestamp]
        :type timestamp: float or :class:`PyTango.TimeVal`
        :param quality:
            attribute quality [default: None, meaning VALID]
        :type quality: :class:`PyTango.AttrQuality`
        :param error:
            a tango DevFailed error or None if not an error [default: None]
        :type error:
            :exc:`PyTango.DevFailed`
        :param priority:
            event priority [default: 1, meaning *normal* priority]. If
            priority is > 1, the event filter is temporarily disabled so the
            event is sent for sure. The event filter is restored to the
            previous value
        :type priority: int
        :param synch:
            If synch is set to True, wait for fire event to finish.
            If False, a job is sent to the sardana thread pool and the method
            returns immediately [default: True]
        """
        set_attr = self.set_attribute_push
        if synch:
            set_attr(attr, value=value, w_value=w_value, timestamp=timestamp,
                     quality=quality, error=error, priority=priority,
                     synch=synch)
        else:
            th_pool = self.get_event_thread_pool()
            th_pool.add(set_attr, None, attr, value=value, w_value=w_value,
                        timestamp=timestamp, quality=quality, error=error,
                        priority=priority, synch=synch)

    def set_attribute_push(self, attr, value=None, w_value=None, timestamp=None,
                           quality=None, error=None, priority=1, synch=True):
        """Synchronous internal implementation of :meth:`set_attribute` (synch
        is passed to this method because it might need to know if it is being
        executed in a synchronous or asynchronous context)."""
        
        if priority > 0 and not synch:
            with self.tango_lock:
                return self._set_attribute_push(attr, value=value,
                        w_value=w_value, timestamp=timestamp, quality=quality,
                        error=error, priority=priority)
        else:
            return self._set_attribute_push(attr, value=value,
                        w_value=w_value, timestamp=timestamp, quality=quality,
                        error=error, priority=priority)

    def _set_attribute_push(self, attr, value=None, w_value=None, timestamp=None,
                            quality=None, error=None, priority=1):
        """Internal method."""
        fire_event = priority > 0

        recover = False
        if priority > 1 and attr.is_check_change_criteria():
            attr.set_change_event(True, False)
            recover = True
        
        attr_name = attr.get_name().lower()

        if value is None and error is None:
            raise Exception("Cannot set value of attribute '%s' with None" % (attr_name,))

        try:
            if error is not None and fire_event:
                self.push_change_event(attr_name, error)
                return

            # some versions of Tango have a memory leak if you do
            # push_change_event(attr_name, value [, ...]) on state or status.
            # This solves the problem.
            if attr_name == "state":
                self.set_state(value)
                if fire_event:
                    self.push_change_event(attr_name)
                return
            elif attr_name == "status":
                self.set_status(value)
                if fire_event:
                    self.push_change_event(attr_name)
                return

            if timestamp is None:
                timestamp = time.time()
            elif isinstance(timestamp, TimeVal):
                timestamp = TimeVal.totime(timestamp)

            if quality is None:
                quality = AttrQuality.ATTR_VALID

            data_type = attr.get_data_type()
            if w_value is not None and isinstance(attr, WAttribute):
                attr.set_write_value(w_value)
            if fire_event:
                if data_type == ArgType.DevEncoded:
                    fmt, data = value
                    args = attr_name, fmt, data, timestamp, quality

                else:
                    args = attr_name, value, timestamp, quality
                self.push_change_event(*args)
            else:
                if data_type == ArgType.DevEncoded:
                    fmt, data = value
                    attr.set_value_date_quality(fmt, data, timestamp, quality)
                else:
                    attr.set_value_date_quality(value, timestamp, quality)
        finally:
            if recover:
                attr.set_change_event(True, True)

    def calculate_tango_state(self, ctrl_state, update=False):
        """Calculate tango state based on the controller state.
        
        :param ctrl_state: the state returned by the controller
        :type ctrl_state: :obj:`~sardana.sardanadefs.State`
        :param bool update:
            if True, set the state of this device with the calculated tango
            state [default: False:
        :return: the corresponding tango state
        :rtype: :class:`PyTango.DevState`"""
        self._state = state = to_tango_state(ctrl_state)
        if update:
            self.set_state(state)
        return state

    def calculate_tango_status(self, ctrl_status, update=False):
        """Calculate tango status based on the controller status.
        
        :param str ctrl_status: the status returned by the controller
        :param bool update:
            if True, set the state of this device with the calculated tango
            state [default: False:
        :return: the corresponding tango state
        :rtype: str"""
        self._status = status = ctrl_status
        if update:
            self.set_status(status)
        return status


class SardanaDeviceClass(DeviceClass):
    """SardanaDeviceClass represents the base class for all Sardana
    :class:`PyTango.DeviceClass` classes"""
    
    #:
    #: Sardana device class properties definition
    #: 
    #: .. seealso:: :ref:`server`
    #:
    class_property_list = {
    }

    #:
    #: Sardana device properties definition
    #: 
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
    }

    #:
    #: Sardana device command definition
    #: 
    #: .. seealso:: :ref:`server`
    #:
    cmd_list = {
    }

    #:
    #: Sardana device attribute definition
    #: 
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
    }

    def __init__(self, name):
        DeviceClass.__init__(self, name)
        self.set_type(name)

    def _get_class_properties(self):
        """Internal method"""
        return dict(ProjectTitle="Sardana", Description="Generic description",
                    doc_url="http://sardana-controls.org/",
                    __icon=self.get_name().lower() + ".png",
                    InheritedFrom=["Device_4Impl"])
    
    def write_class_property(self):
        """Write class properties ``ProjectTitle``, ``Description``, 
        ``doc_url``, ``InheritedFrom`` and ``__icon``"""
        db = self.get_database()
        if db is None:
            return
        db.put_class_property(self.get_name(), self._get_class_properties())

    def dyn_attr(self, dev_list):
        """Invoked to create dynamic attributes for the given devices.
        Default implementation calls
        :meth:`SardanaDevice.initialize_dynamic_attributes` for each device
        
        :param dev_list: list of devices
        :type dev_list: :class:`PyTango.DeviceImpl`"""
        for dev in dev_list:
            try:
                dev.initialize_dynamic_attributes()
            except:
                dev.warning("Failed to initialize dynamic attributes")
                dev.debug("Details:", exc_info=1)

    def device_name_factory(self, dev_name_list):
        """Builds list of device names to use when no Database is being used
        
        :param dev_name_list: list to be filled with device names
        :type dev_name_list: seq<obj:`list`\>"""
        tango_class = self.get_name()
        devices = NO_DB_MAP.get(tango_class, ())
        for dev_info in devices:
            dev_name_list.append(dev_info[1])



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the
PoolPseudoCounter class"""

__all__ = [ "PoolPseudoCounter" ]

__docformat__ = 'restructuredtext'

import sys
import time

from sardana import State, ElementType, TYPE_PHYSICAL_ELEMENTS
from sardana.sardanaattribute import SardanaAttribute
from sardana.sardanaexception import SardanaException
from sardana.sardanavalue import SardanaValue

from .poolexception import PoolException
from .poolbasechannel import PoolBaseChannel
from .poolbasegroup import PoolBaseGroup
from .poolacquisition import PoolAcquisition


class Value(SardanaAttribute):
    
    def __init__(self, *args, **kwargs):
        self._exc_info = None
        super(Value, self).__init__(*args, **kwargs)
        for value_attr in self.obj.get_physical_value_attribute_iterator():
            value_attr.add_listener(self.on_change)
    
    def _in_error(self):
        for value_attr in self.obj.get_physical_value_attribute_iterator():
            if value_attr.error:
                return True
        return self._exc_info != None
    
    def _has_value(self):
        for value_attr in self.obj.get_physical_value_attribute_iterator():
            if not value_attr.has_value():
                return False
        return True
    
    def _get_value(self):
        return self.calc().value

    def _set_value(self, value, exc_info=None, timestamp=None, propagate=1):
        raise Exception("Cannot set value for %s" % self.obj.name)

    def _get_write_value(self):
        w_values = self.get_physical_write_values()
        return self.calc_pseudo(physical_values=w_values).value
    
    def _set_write_value(self, w_value, timestamp=None, propagate=1):
        raise Exception("Cannot set write value for %s" % self.obj.name)
    
    def _get_exc_info(self):
        exc_info = self._exc_info
        if exc_info is None:
            for value_attr in self.obj.get_physical_value_attribute_iterator():
                if value_attr.error:
                    return value_attr.get_exc_info()
        return exc_info
    
    def _get_timestamp(self):
        timestamps = [ value_attr.timestamp for value_attr in self.obj.get_physical_value_attribute_iterator() ]
        if not len(timestamps):
            timestamps = self._local_timestamp,
        return max(timestamps)

    def get_physical_write_values(self):
        ret = []
        for value_attr in self.obj.get_physical_value_attribute_iterator():
            if value_attr.has_write_value():
                value = value_attr.w_value
            else:
                if not value_attr.has_value():
                    # if underlying counter doesn't have value yet, it is
                    # because of a cold start
                    value_attr.update(propagate=0)
                if value_attr.in_error():
                    raise PoolException("Cannot get '%' value" % value_attr.obj.name,
                                        exc_info=value_attr.exc_info)
                value = value_attr.value
            ret.append(value)
        return ret
        
    def get_physical_values(self):
        ret = []
        for value_attr in self.obj.get_physical_value_attribute_iterator():
            # if underlying channel doesn't have value yet, it is because
            # of a cold start
            if not value_attr.has_value():
                value_attr.update(propagate=0)
            if value_attr.in_error():
                raise PoolException("Cannot get '%' value" % value_attr.obj.name,
                                    exc_info=value_attr.exc_info)
            ret.append(value_attr.value)
        return ret

    def calc(self, physical_values=None):
        try:
            obj = self.obj
            if physical_values is None:
                physical_values = self.get_physical_values()
            else:
                l_v, l_u = len(physical_values), len(obj.get_user_elements())
                if l_v != l_u:
                    raise IndexError("CalcPseudo(%s): must give %d physical " \
                                     "values (you gave %d)" % (obj.name, l_u, l_v) )
            ctrl, axis = obj.controller, obj.axis    
            result = ctrl.calc(axis, physical_values)
        except SardanaException as se:
            result = SardanaValue(exc_info=se.exc_info)
        except:
            result = SardanaValue(exc_info=sys.exc_info())
        return result    

    def calc_all(self, physical_values=None):
        try:
            obj = self.obj
            if physical_values is None:
                physical_values = self.get_physical_values()
            else:
                l_v, l_u = len(physical_values), len(obj.get_user_elements())
                if l_v != l_u:
                    raise IndexError("CalcAllPseudo(%s): must give %d physical " \
                                     "values (you gave %d)" % (obj.name, l_u, l_v) )
            ctrl, axis = obj.controller, obj.axis    
            result = ctrl.calc_all(axis, physical_values)
        except SardanaException as se:
            result = SardanaValue(exc_info=se.exc_info)
        except:
            result = SardanaValue(exc_info=sys.exc_info())
        return result   
    
    def on_change(self, evt_src, evt_type, evt_value):
        self.fire_read_event(propagate=evt_type.priority)

    def update(self, cache=True, propagate=1):
        if cache:
            for phy_elem_val in self.obj.get_low_level_physical_value_attribute_iterator():
                if not phy_elem_val.has_value():
                    cache = False
                    break
        if not cache:
            values = self.obj.acquisition.read_value(serial=True)
            if not len(values):
                self._local_timestamp = time.time()
            for acq_obj, value in values.items():
                acq_obj.put_value(value, propagate=propagate)


class PoolPseudoCounter(PoolBaseGroup, PoolBaseChannel):
    """A class representing a Pseudo Counter in the Sardana Device Pool"""
    
    ValueAttributeClass = Value
    AcquisitionClass = None
    
    def __init__(self, **kwargs):
        self._siblings = None
        user_elements = kwargs.pop('user_elements')
        kwargs['elem_type'] = ElementType.PseudoCounter
        # don't switch the order of constructors!
        PoolBaseGroup.__init__(self, user_elements=user_elements,
                               pool=kwargs['pool'])
        PoolBaseChannel.__init__(self, **kwargs)

    def serialize(self, *args, **kwargs):
        kwargs = PoolBaseChannel.serialize(self, *args, **kwargs)
        elements = [ elem.name for elem in self.get_user_elements() ]
        physical_elements = []
        for elem_list in self.get_physical_elements().values():
            for elem in elem_list:
                physical_elements.append(elem.name)
        cl_name = self.__class__.__name__
        cl_name = cl_name[4:]
        kwargs['elements'] = elements
        kwargs['physical_elements'] = physical_elements
        return kwargs

    def on_element_changed(self, evt_src, evt_type, evt_value):
        name = evt_type.name.lower()
        # always calculate state.
        status_info = self._calculate_states()
        state, status = self.calculate_state_info(status_info=status_info)
        state_propagate = 0
        status_propagate = 0
        if name == 'state':
            state_propagate = evt_type.priority
        elif name == 'status':
            status_propagate = evt_type.priority
        self.set_state(state, propagate=state_propagate)
        self.set_status(status, propagate=status_propagate)
    
    def _create_action_cache(self):
        acq_name = "%s.Acquisition" % self._name
        return PoolAcquisition(self, acq_name)
    
    def get_action_cache(self):
        return self._get_action_cache()
    
    def set_action_cache(self, action_cache):
        self._set_action_cache(action_cache)

    def get_siblings(self):
        if self._siblings is None:
            self._siblings = siblings = set()
            for axis, sibling in self.controller.get_element_axis().items():
                if axis == self.axis:
                    continue
                siblings.add(sibling)
        return self._siblings

    siblings = property(fget=get_siblings,
                        doc="the siblings for this pseudo counter")
    
    # ------------------------------------------------------------------------
    # value
    # ------------------------------------------------------------------------
    
    def calc(self, physical_values=None):
        return self.get_value_attribute().calc(physical_values=physical_values)    

    def calc_all(self, physical_values=None):
        return self.get_value_attribute().calc_all(physical_values=physical_values)    
    
    def get_low_level_physical_value_attribute_iterator(self):
        return self.get_physical_elements_attribute_iterator()
        
    def get_physical_value_attribute_iterator(self):
        return self.get_user_elements_attribute_iterator()
    
    def get_physical_values_attribute_sequence(self):
        return self.get_user_elements_attribute_sequence()
    
    def get_physical_values_attribute_map(self):
        return self.get_user_elements_attribute_map()
    
    def get_physical_values(self, cache=True, propagate=1):
        """Get value for underlying elements.

        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the physical value
        :rtype:
            dict <PoolElement, :class:`~sardana.sardanaattribute.SardanaAttribute` >"""
        self._value.update(cache=cache, propagate=propagate)
        return self.get_physical_values_attribute_map()

    def get_siblings_values(self, use=None):
        """Get the last values for all siblings.
        
        :param use: the already calculated values. If a sibling is in this
                    dictionary, the value stored here is used instead
        :type use: dict <PoolElement, :class:`~sardana.sardanavalue.SardanaValue` >
        :return: a dictionary with siblings values
        :rtype:
            dict <PoolElement, value(float?) >"""        
        values = {}
        for sibling in self.siblings:
            value_attr = sibling.get_value_attribute() 
            if use and sibling in use:
                pos = use[sibling]
            else:
                pos = value_attr.value
            values[sibling] = pos
        return values
    
    def get_value(self, cache=True, propagate=1):
        """Returns the pseudo counter value.
        
        :param cache:
            if ``True`` (default) return value in cache, otherwise read value
            from hardware
        :type cache:
            bool
        :param propagate:
            0 for not propagating, 1 to propagate, 2 propagate with priority
        :type propagate:
            int
        :return:
            the pseudo counter value
        :rtype:
            :class:`~sardana.sardanaattribute.SardanaAttribute`"""
        value_attr = self._value
        value_attr.update(cache=cache, propagate=propagate)
        return value_attr
    
    def set_value(self, value, propagate=1):
        raise Exception("Not possible to set_value of a pseudo counter")
        
    value = property(get_value, doc="pseudo counter value")

    # --------------------------------------------------------------------------
    # state information
    # --------------------------------------------------------------------------

    _STD_STATUS = "{name} is {state}\n{ctrl_status}"
    def calculate_state_info(self, status_info=None):
        if status_info is None:
            status_info = self._state, self._status
        state, status = status_info
        if state == State.On:
            state_str = "Stopped"
        else:
            state_str = "in " + State[state]
        new_status = self._STD_STATUS.format(name=self.name, state=state_str,
                                             ctrl_status=status)
        return status_info[0], new_status
    
    def read_state_info(self, state_info=None):
        if state_info is None:
            state_info = {}
            action_cache = self.get_action_cache()
            ctrl_state_infos = action_cache.read_state_info(serial=True)
            for obj, ctrl_state_info in ctrl_state_infos.items():
                state_info[obj] = state_info = \
                    obj._from_ctrl_state_info(ctrl_state_info)
                obj.put_state_info(state_info)
        for user_element in self.get_user_elements():
            if user_element.get_type() not in TYPE_PHYSICAL_ELEMENTS:
                state_info = user_element._calculate_states()
                user_element.put_state_info(state_info)
        
        ret = self._calculate_states()
        return ret




##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Environment related macros"""

__all__ = ["dumpenv", "load_env", "lsenv", "senv", "usenv"]

__docformat__ = 'restructuredtext'

from taurus.console.list import List
from sardana.macroserver.macro import *

################################################################################
#
# Environment related macros
#
################################################################################

from lxml import etree

def reprValue(v, max=74):
    # cut long strings
    v = str(v)
    if len(v) > max:
        v = v[:max] + ' [...]'
    return v


class dumpenv(Macro):
    """Dumps the complete environment"""
    
    def run(self):
        env = self.getGlobalEnv()
        out = List(['Name','Value','Type'])
        for k,v in env.iteritems():
            str_v = reprValue(v)
            type_v = type(v).__name__
            out.appendRow([str(k), str_v, type_v])

        for line in out.genOutput():
            self.output(line)


class lsvo(Macro):
    """Lists the view options"""
    def run(self):
        vo = self.getViewOptions()
        out = List(['View option', 'Value'])
        for key, value in vo.items():
            out.appendRow([key, str(value)])

        for line in out.genOutput():
            self.output(line)


class setvo(Macro):
    """Sets the given view option to the given value"""

    param_def = [['name', Type.String, None, 'View option name'],
                 ['value', Type.String, None, 'View option value']]

    def run(self, name, value):
        try:
            value = eval(value)
        except:
            pass
        self.setViewOption(name, value)


class usetvo(Macro):
    """Resets the value of the given view option"""
    
    param_def = [['name', Type.String, None, 'View option name']]

    def run(self, name):
        self.resetViewOption(name)

            
class lsenv(Macro):
    """Lists the environment"""
    
    param_def = [
        ['macro_list',
         ParamRepeat(['macro', Type.MacroClass, None, 'macro name'], min=0),
         None, 'List of macros to show environment'],
    ]
    
    def prepare(self, *macro_list, **opts):
        self.table_opts = opts
        
    def run(self, *macro_list):
        # list the environment for the current door
        if len(macro_list) == 0:
            # list All the environment for the current door
            out = List(['Name', 'Value', 'Type'])
            env = self.getAllDoorEnv()
            for k,v in env.iteritems():
                str_val = reprValue(v)
                type_name = type(v).__name__
                out.appendRow([str(k), str_val, type_name])
        else:
            # list the environment for the current door for the given macros
            out = List(['Macro', 'Name', 'Value', 'Type'])
            for macro in macro_list:
                env = self.getEnv(key=None, macro_name=macro.name)
                for k, v in env.iteritems():
                    type_name = type(v).__name__
                    out.appendRow([ macro.name, k, self.reprValue(v), type_name ])

        for line in out.genOutput():
            self.output(line)

    def reprValue(self, v, max=54):
        # cut long strings
        v = str(v)
        if len(v) > max: v = '%s [...]' % v[:max]
        return v

class senv(Macro):
    """Sets the given environment variable to the given value"""

    param_def = [['name', Type.Env, None,
                  'Environment variable name. Can be one of the following:\n' \
                  ' - <name> - global variable\n' \
                  ' - <full door name>.<name> - variable value for a specific door\n' \
                  ' - <macro name>.<name> - variable value for a specific macro\n' \
                  ' - <full door name>.<macro name>.<name> - variable value for a specific macro running on a specific door'],
                 ['value_list',
                  ParamRepeat(['value', Type.String, None, 'environment value item'], min=1),
                  None, 'value(s). one item will eval to a single element. More than one item will eval to a tuple of elements'],
                ]

    def run(self, env, *value):
        if len(value) == 1: 
            value = value[0]
        else:
            value = '(%s)' % ', '.join(value)
        k,v = self.setEnv(env, value)
        line = '%s = %s' % (k, str(v))
        self.output(line)

class usenv(Macro):
    """Unsets the given environment variable"""
    param_def = [
        ['environment_list',
         ParamRepeat(['env', Type.Env, None, 'Environment variable name'], min=1),
         None, 'List of environment items to be removed'],
    ]    
    
    def run(self, *env):
        self.unsetEnv(env)
        self.output("Success!")
        
class load_env(Macro):
    """ Read environment variables from config_env.xml file"""
    
    def run(self):
        doc = etree.parse("config_env.xml")       
        root = doc.getroot()
        for element in root:
            if element.find("./name").text == "auto_filter":
                self.output("Loading auto_filter variables:")
                filter_max_elem = element.find(".//FilterMax")
                if filter_max_elem is not None:
                    filter_max = filter_max_elem.text
                    self.setEnv("FilterMax", filter_max)
                    self.output("FilterMax loaded")
                else:
                    self.output("FilterMax not found")
                filter_min_elem = element.find(".//FilterMin")
                if filter_min_elem is not None:
                    filter_min = filter_max_elem.text
                    self.setEnv("FilterMin", filter_min)
                    self.output("FilterMin loaded")
                else:
                    self.output("FilterMin not found")
                filter_delta_elem = element.find(".//FilterDelta")
                if filter_delta_elem is not None:
                    filter_delta = filter_delta_elem.text
                    self.setEnv("FilterDelta", filter_delta)
                    self.output("FilterDelta loaded")
                else:
                    self.output("FilterDelta not found")
                filter_signal_elem = element.find(".//FilterSignal")
                if filter_signal_elem is not None:
                    filter_signal = filter_signal_elem.text
                    self.setEnv("FilterSignal", filter_signal)
                    self.output("FilterSignal loaded")
                else:
                    self.output("FilterSignal not found")
                filter_absorber_elem = element.find(".//FilterAbsorber")
                if filter_absorber_elem is not None:
                    filter_absorber = filter_absorber_elem.text
                    self.setEnv("FilterAbsorber", filter_absorber)
                    self.output("FilterAbsorber loaded")
                else:
                    self.output("FilterAbsorber not found")
                auto_filter_elem = element.find(".//AutoFilter")
                if auto_filter_elem is not None:
                    auto_filter = auto_filter_elem.text
                    self.setEnv("AutoFilter", auto_filter)
                    self.output("AutoFilter loaded")
                else:
                    self.output("AutoFilter not found")
            if element.find("./name").text == "auto_beamshutter":
                self.output("Loading auto_beamshutter variables:")
                auto_beamshutter_elem = element.find(".//AutoBeamshutter")
                if auto_beamshutter_elem is not None:
                    auto_beamshutter = auto_beamshutter_elem.text
                    self.setEnv("AutoBeamshutter", auto_beamshutter)
                    self.output("AutoBeamshutter loaded")
                else:
                    self.output("AutoBeamshutter not found")
                beamshutter_limit_elem = element.find(".//BeamshutterLimit")
                if beamshutter_limit_elem is not None:
                    beamshutter_limit = beamshutter_limit_elem.text
                    self.setEnv("BeamshutterLimit", beamshutter_limit)
                    self.output("BeamshutterLimit loaded")
                else:
                    self.output("BeamshutterLimit not found")
                beamshutter_signal_elem = element.find(".//BeamshutterSignal")
                if beamshutter_signal_elem is not None:
                    beamshutter_signal = beamshutter_signal_elem.text
                    self.setEnv("BeamshutterSignal", beamshutter_signal)
                    self.output("BeamshutterSignal loaded")
                else:
                    self.output("BeamshutterSignal not found")
                beamshutter_time_elem = element.find(".//BeamshutterTime")
                if beamshutter_time_elem is not None:
                    beamshutter_time = beamshutter_time_elem.text
                    self.setEnv("BeamshutterTime", beamshutter_time)
                    self.output("BeamshutterTime loaded")
                else:
                    self.output("BeamshutterTime not found")
            if element.find("./name").text == "exafs":
                self.output("Loading exafs variables:")
                exafs_int_times_elem = element.find(".//ExafsIntTimes")
                if exafs_int_times_elem is not None:
                    exafs_int_times = exafs_int_times_elem.text
                    self.setEnv("ExafsIntTimes", exafs_int_times)
                    self.output("ExafsIntTimes loaded")
                else:
                    self.output("ExafsIntTimes not found")
                exafs_nb_intervals_elem = element.find(".//ExafsNbIntervals")
                if exafs_nb_intervals_elem is not None:
                    exafs_nb_intervals = exafs_nb_intervals_elem.text
                    self.setEnv("ExafsNbIntervals", exafs_nb_intervals)
                    self.output("ExafsNbIntervals loaded")
                else:
                    self.output("ExafsNbIntervals not found")
                exafs_regions_elem = element.find(".//ExafsRegions")
                if exafs_regions_elem is not None:
                    exafs_regions = exafs_regions_elem.text
                    self.setEnv("ExafsRegions", exafs_regions)
                    self.output("ExafsRegions loaded")
                else:
                    self.output("ExafsRegions not found")  
        misc_tree = root.find("./miscellaneous")
        if misc_tree is not None:
            for parameter in misc_tree:
                if parameter.tag != "name":
                    self.setEnv(parameter.tag, parameter.text) 
            



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for a pool container element"""

__all__ = ["SardanaContainer"]

__docformat__ = 'restructuredtext'

from taurus.core.util import CaselessDict
from sardanadefs import InvalidId, ElementType

class SardanaContainer(object):
    """A container class for sardana elements"""
    
    def __init__(self):
        
        # map of all elements
        # key - element ID
        # value - pointer to the element object
        self._element_ids = {}

        # map of all elements by name
        # key - element name
        # value - pointer to the element object
        self._element_names = CaselessDict()

        # map of all elements by name
        # key - element full name
        # value - pointer to the element object
        self._element_full_names = CaselessDict()
    
        # map of all elements by type
        # key - element type
        # value - map where:
        #    key - element ID
        #    value - pointer to the element object
        self._element_types = {}
        
    def add_element(self, e):
        """Adds a new :class:`pool.PoolObject` to this container
           
           :param e: the pool element to be added
           :type e: :class:`pool.PoolObject`
        """
        name, full_name, id = e.get_name(), e.get_full_name(), e.get_id()
        elem_type = e.get_type()
        self._element_ids[id] = e
        self._element_names[name] = e
        self._element_full_names[full_name] = e
        type_elems = self._element_types.get(elem_type)
        if type_elems is None:
            self._element_types[elem_type] = type_elems = {}
        type_elems[id] = e
        return e
    
    def remove_element(self, e):
        """Removes the :class:`pool.PoolObject` from this container
           
           :param e: the pool object to be removed
           :type e: :class:`pool.PoolObject`
           
           :throw: KeyError
        """
        name, full_name, id = e.get_name(), e.get_full_name(), e.get_id()
        elem_type = e.get_type()
        del self._element_ids[id]
        del self._element_names[name]
        del self._element_full_names[full_name]
        type_elems = self._element_types.get(elem_type)
        del type_elems[id]
        
    def get_element_id_map(self):
        """Returns a reference to the internal pool object ID map
           
           :return: the internal pool object ID map
           :rtype: dict<id, pool.PoolObject>
        """
        return self._element_ids
    
    def get_element_name_map(self):
        """Returns a reference to the internal pool object name map
           
           :return: the internal pool object name map
           :rtype: dict<str, pool.PoolObject>
        """
        return self._element_names
    
    def get_element_type_map(self):
        """Returns a reference to the internal pool object type map
           
           :return: the internal pool object type map
           :rtype: dict<pool.ElementType, dict<id, pool.PoolObject>>
        """
        return self._element_types
    
    def get_element(self, **kwargs):
        """Returns a reference to the requested pool object
           
           :param kwargs: if key 'id' given: search by ID
                          else if key 'full_name' given: search by full name
                          else if key 'name' given: search by name
           
           :return: the pool object 
           :rtype: pool.PoolObject
           
           :throw: KeyError
        """
        if kwargs.has_key("id"):
            id = kwargs.pop("id")
            return self.get_element_by_id(id, **kwargs)
        
        if kwargs.has_key("full_name"):
            full_name = kwargs.pop("full_name")
            return self.get_element_by_full_name(full_name, **kwargs)
        
        name = kwargs.pop("name")
        return self.get_element_by_name(name, **kwargs)
    
    def get_element_by_name(self, name, **kwargs):
        """Returns a reference to the requested pool object
           
           :param name: pool object name
           :type name: str
           
           :return: the pool object 
           :rtype: pool.PoolObject
           
           :throw: KeyError
        """
        ret = self._element_names.get(name)
        if ret is None:
            raise KeyError("There is no element with name '%s'" % name)
        return ret
    
    def get_element_by_full_name(self, full_name, **kwargs):
        """Returns a reference to the requested pool object
           
           :param name: pool object full name
           :type name: str
           
           :return: the pool object 
           :rtype: pool.PoolObject
           
           :throw: KeyError
        """
        ret = self._element_full_names.get(full_name)
        if ret is None: 
            raise KeyError("There is no element with full name '%s'" % full_name)
        return ret
        
    def get_element_by_id(self, id, **kwargs):
        """Returns a reference to the requested pool object
           
           :param id: pool object ID
           :type id: int
           
           :return: the pool object 
           :rtype: pool.PoolObject
           
           :throw: KeyError
        """
        ret = self._element_ids.get(id)
        if ret is None: 
            raise KeyError("There is no element with ID '%d'" % id)
        return ret
    
    def get_elements_by_type(self, t):
        """Returns a list of all pool objects of the given type
           
           :param t: element type
           :type t: pool.ElementType
           
           :return: list of pool objects
           :rtype: seq<pool.PoolObject>
        """
        elem_types_dict = self._element_types.get(t)
        if elem_types_dict is None:
            return  []
        return elem_types_dict.values()

    def get_element_names_by_type(self, t):
        """Returns a list of all pool object names of the given type
           
           :param t: element type
           :type t: pool.ElementType
           
           :return: list of pool object names
           :rtype: seq<str>
        """
        return [ elem.get_name() for elem in self.get_elements_by_type(t) ]
        
    def rename_element(self, old_name, new_name):
        """Rename a pool object
           
           :param old_name: old pool object name
           :type old_name: str
           :param new_name: new pool object name
           :type new_name: str
        """
        raise NotImplementedError

    def check_element(self, name, full_name):
        raise_element_name = True
        try:
            elem = self.get_element(name=name)
        except:
            raise_element_name = False
        if raise_element_name:
            elem_type = ElementType[elem.get_type()]
            raise Exception("A %s with name '%s' already exists"
                            % (elem_type, name))
        
        raise_element_full_name = True
        try:
            elem = self.get_element(full_name=full_name)
        except:
            raise_element_full_name = False
        if raise_element_full_name:
            elem_type = ElementType[elem.get_type()]
            raise Exception("A %s with full name '%s' already exists"
                            % (elem_type, full_name))


##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
    Macro library containning examples demonstrating specific features or tricks
    for programming macros for Sardana.
   
   Available Macros are:
     ascanr
     toothedtriangle
     
"""

__all__ = ["ascan_demo", "ascanr", "toothedtriangle", "regscan", "reg2scan", "reg3scan", "a2scan_mod"]

__docformat__ = 'restructuredtext'

import os
import numpy

from sardana.macroserver.macro import *
from sardana.macroserver.scan import *

class ascan_demo(Macro):
    """
    This is a basic reimplementation of the ascan` macro for demonstration
    purposes of the Generic Scan framework. The "real" implementation of
    :class:`sardana.macroserver.macros.ascan` derives from
    :class:`sardana.macroserver.macros.aNscan` and provides some extra features.
    """

    hints = { 'scan' : 'ascan_demo'} #this is used to indicate other codes that the macro is a scan
    env = ('ActiveMntGrp',) #this hints that the macro requires the ActiveMntGrp environment variable to be set

    param_def = [
       ['motor',      Type.Moveable, None, 'Motor to move'],
       ['start_pos',  Type.Float,    None, 'Scan start position'],
       ['final_pos',  Type.Float,    None, 'Scan final position'],
       ['nr_interv',  Type.Integer,  None, 'Number of scan intervals'],
       ['integ_time', Type.Float,    None, 'Integration time']
    ]

    def prepare(self, motor, start_pos, final_pos, nr_interv, integ_time, **opts):
        #parse the user parameters
        self.start = numpy.array([start_pos], dtype='d')
        self.final = numpy.array([final_pos], dtype='d')
        self.integ_time = integ_time

        self.nr_points = nr_interv+1
        self.interv_size = ( self.final - self.start) / nr_interv
        self.name='ascan_demo'
        env = opts.get('env',{}) #the "env" dictionary may be passed as an option
        
        #create an instance of GScan (in this case, of its child, SScan
        self._gScan=SScan(self, generator=self._generator, moveables=[motor], env=env) 
  

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time #integ_time is the same for all steps
        for point_no in xrange(self.nr_points):
            step["positions"] = self.start + point_no * self.interv_size #note that this is a numpy array
            step["point_id"] = point_no
            yield step
    
    def run(self,*args):
        for step in self._gScan.step_scan(): #just go through the steps
            yield step
        
    @property
    def data(self):
        return self._gScan.data #the GScan provides scan data


class ascanr(Macro, Hookable):
    """This is an example of how to handle adding extra info columns in a scan.
    Does the same than ascan but repeats the acquisitions "repeat" times for each step. 
    It could be implemented deriving from aNscan, but I do it like this for clarity.
    Look for the comments with "!!!" for tips specific to the extra info columns
    I do not support constrains in this one for simplicity (see ascan for that)
    
    Do an absolute scan of the specified motor, repeating measurements in each step.
    ascanr scans one motor, as specified by motor. The motor starts at the
    position given by start_pos and ends at the position given by final_pos.
    At each step, the acquisition will be repeated "repeat" times
    The step size is (start_pos-final_pos)/nr_interv. The number of data points collected
    will be (nr_interv+1)*repeat. Count time for each acquisition is given by time which if positive,
    specifies seconds and if negative, specifies monitor counts. """

    hints = { 'scan' : 'ascanr', 'allowsHooks':('pre-move', 'post-move', 'pre-acq', 'post-acq', 'post-step') }
    env = ('ActiveMntGrp',)

    param_def = [
       ['motor',      Type.Moveable, None, 'Motor to move'],
       ['start_pos',  Type.Float,    None, 'Scan start position'],
       ['final_pos',  Type.Float,    None, 'Scan final position'],
       ['nr_interv',  Type.Integer,  None, 'Number of scan intervals'],
       ['integ_time', Type.Float,    None, 'Integration time'],
       ['repeat',     Type.Integer,  None, 'Number of Repetitions']
    ]


    def prepare(self, motor, start_pos, final_pos, nr_interv, integ_time, repeat,
                **opts):
        
        self.starts = numpy.array([start_pos], dtype='d')
        self.finals = numpy.array([final_pos], dtype='d')
        self.nr_interv = nr_interv
        self.integ_time = integ_time
        self.repeat=repeat
        self.opts = opts

        self.nr_points = nr_interv+1
        self.interv_sizes = ( self.finals - self.starts) / nr_interv
        self.name='ascanr'
        
        generator=self._generator
        moveables=[motor]
        env=opts.get('env',{})
        constrains=[]
        extrainfodesc=[ColumnDesc(name='repetition',
                                  dtype='int64', shape=(1,))] #!!!
                
        self._gScan=SScan(self, generator, moveables, env, constrains, extrainfodesc) #!!!
  

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time
        step["post-acq-hooks"] = self.getHooks('post-acq') +  self.getHooks('_NOHINT_')
        step["post-step-hooks"] = self.getHooks('post-step')
        step["check_func"] = []
        extrainfo = {"repetition":0} #!!! 
        step['extrainfo'] = extrainfo  #!!! 
        for point_no in xrange(self.nr_points):
            step["positions"] = self.starts + point_no * self.interv_sizes
            step["point_id"] = point_no
            for i in xrange(self.repeat):
                extrainfo["repetition"] = i #!!! 
                yield step
    
    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step
        
    @property
    def data(self):
        return self._gScan.data
        

class toothedtriangle(Macro, Hookable):
    """toothedtriangle macro implemented with the gscan framework.
    It performs nr_cycles cycles, each consisting of two stages: the first half
    of the cycle it behaves like the ascan macro (from start_pos to stop_pos in
    nr_interv+1 steps).For the second half of the cycle it steps back until
    it undoes the first half and is ready for the next cycle.
    At each step, nr_samples acquisitions are performed.
    The total number of points in the scan is nr_interv*2*nr_cycles*nr_samples+1"""

    hints = { 'scan' : 'toothedtriangle',
             'allowsHooks':('pre-scan', 'pre-move', 'post-move', 'pre-acq',
                            'post-acq', 'post-step', 'post-scan')
             }
    env = ('ActiveMntGrp',)

    param_def = [
       ['motor',      Type.Moveable, None, 'Motor to move'],
       ['start_pos',  Type.Float,    None, 'start position'],
       ['final_pos',  Type.Float,    None, 'position after half cycle'],
       ['nr_interv',  Type.Integer,  None, 'Number of intervals in half cycle'],
       ['integ_time', Type.Float,    None, 'Integration time'],
       ['nr_cycles',  Type.Integer,  None, 'Number of cycles'],
       ['nr_samples', Type.Integer,  1 , 'Number of samples at each point']
    ]

    def prepare(self, motor, start_pos, final_pos, nr_interv, integ_time,
                nr_cycles, nr_samples, **opts):
        
        self.start_pos = start_pos
        self.final_pos = final_pos
        self.nr_interv = nr_interv
        self.integ_time = integ_time
        self.nr_cycles = nr_cycles
        self.nr_samples = nr_samples
        self.opts = opts
        cycle_nr_points = self.nr_interv+1 + (self.nr_interv+1)-2
        self.nr_points = cycle_nr_points*nr_samples*nr_cycles+nr_samples
        
        self.interv_size = ( self.final_pos - self.start_pos) / nr_interv
        self.name='toothedtriangle'
        
        generator=self._generator
        moveables = []
        moveable = MoveableDesc(moveable=motor, is_reference=True,
                                min_value=min(start_pos,final_pos),
                                max_value=max(start_pos,final_pos))
        moveables=[moveable]
        env=opts.get('env',{})
        constrains=[]
        extrainfodesc=[ColumnDesc(name='cycle', dtype='int64', shape=(1,)),
                       ColumnDesc(name='interval', dtype='int64', shape=(1,)),
                       ColumnDesc(name='sample', dtype='int64', shape=(1,))] #!!!
                
        self._gScan=SScan(self, generator, moveables, env, constrains, extrainfodesc) #!!!
  

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time
        step["pre-move-hooks"] = self.getHooks('pre-move')
        step["post-move-hooks"] = self.getHooks('post-move')
        step["pre-acq-hooks"] = self.getHooks('pre-acq')
        step["post-acq-hooks"] = self.getHooks('post-acq') + self.getHooks('_NOHINT_')
        step["post-step-hooks"] = self.getHooks('post-step')
        step["check_func"] = []
        extrainfo = {"cycle":None, "interval":None, "sample":None, } 
        step['extrainfo'] = extrainfo
        halfcycle1=range(self.nr_interv+1)
        halfcycle2=halfcycle1[1:-1]
        halfcycle2.reverse()
        intervallist=halfcycle1+halfcycle2
        point_no=0
        for cycle in xrange(self.nr_cycles):
            extrainfo["cycle"] = cycle
            for interval in intervallist:
                extrainfo["interval"] = interval
                step["positions"] = numpy.array([self.start_pos + (interval) * self.interv_size] ,dtype='d')
                for sample in xrange(self.nr_samples):
                    extrainfo["sample"] = sample
                    step["point_id"] = point_no 
                    yield step
                    point_no+=1
                    
        #last step for closing the loop
        extrainfo["interval"] = 0
        step["positions"] = numpy.array([self.start_pos] ,dtype='d')
        for sample in xrange(self.nr_samples):
            extrainfo["sample"] = sample
            step["point_id"] = point_no 
            yield step
            point_no+=1
    
    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step
    
    @property
    def data(self):
        return self._gScan.data


class regscan(Macro):
    """regscan.
    Do an absolute scan of the specified motor with different number of intervals for each region.
    It uses the gscan framework.

    NOTE: Due to a ParamRepeat limitation, integration time has to be
    specified before the regions.
    """

    hints = {'scan' : 'regscan'}
    env = ('ActiveMntGrp',)

    param_def = [
        ['motor',      Type.Moveable, None, 'Motor to move'],
        ['integ_time', Type.Float,    None, 'Integration time'],
        ['start_pos',  Type.Float,    None, 'Start position'],
        ['step_region',
         ParamRepeat(['next_pos',  Type.Float,   None, 'next position'],
                     ['region_nr_intervals',  Type.Float,   None, 'Region number of intervals']),
         None, 'List of tuples: (next_pos, region_nr_intervals']
    ]

    def prepare(self, motor, integ_time, start_pos, *regions, **opts):
        self.name='regscan'
        self.integ_time = integ_time
        self.start_pos = start_pos
        self.regions = regions
        self.regions_count = len(self.regions)/2
        
        generator=self._generator
        moveables=[motor]
        env=opts.get('env',{})
        constrains=[]
        self._gScan=SScan(self, generator, moveables, env, constrains)

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time
        
        point_id = 0
        region_start = self.start_pos
        for r in range(len(self.regions)):
            region_stop, region_nr_intervals = self.regions[r][0], self.regions[r][1]
            positions = numpy.linspace(region_start, region_stop, region_nr_intervals+1)
            if region_start != self.start_pos:
                # positions must be calculated from the start to the end of the region
                # but after the first region, the 'start' point must not be repeated
                positions = positions[1:]
            for p in positions:
                step['positions'] = [p]
                step['point_id'] = point_id
                point_id += 1
                yield step
            region_start = region_stop

    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step


class reg2scan(Macro):
    """reg2scan.
    Do an absolute scan of the specified motors with different number of intervals for each region.
    It uses the gscan framework. All the motors will be drived to the same position in each step

    NOTE: Due to a ParamRepeat limitation, integration time has to be
    specified before the regions.
    """

    hints = {'scan' : 'reg2scan'}
    env = ('ActiveMntGrp',)

    param_def = [
        ['motor1',     Type.Moveable, None, 'Motor to move'],
        ['motor2',     Type.Moveable, None, 'Motor to move'],
        ['integ_time', Type.Float,    None, 'Integration time'],
        ['start_pos',  Type.Float,    None, 'Start position'],
        ['step_region',
         ParamRepeat(['next_pos',  Type.Float,   None, 'next position'],
                     ['region_nr_intervals',  Type.Float,   None, 'Region number of intervals']),
         None, 'List of tuples: (next_pos, region_nr_intervals']
    ]

    def prepare(self, motor1, motor2, integ_time, start_pos, *regions, **opts):
        self.name='reg2scan'
        self.integ_time = integ_time
        self.start_pos = start_pos
        self.regions = regions
        self.regions_count = len(self.regions)/2

        generator=self._generator
        moveables=[motor1, motor2]
        env=opts.get('env',{})
        constrains=[]
        self._gScan=SScan(self, generator, moveables, env, constrains)

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time

        point_id = 0
        region_start = self.start_pos
        for r in range(len(self.regions)):
            region_stop, region_nr_intervals = self.regions[r][0], self.regions[r][1]
            positions = numpy.linspace(region_start, region_stop, region_nr_intervals+1)
            if region_start != self.start_pos:
                # positions must be calculated from the start to the end of the region
                # but after the first region, the 'start' point must not be repeated
                positions = positions[1:]
            for p in positions:
                step['positions'] = [p, p]
                step['point_id'] = point_id
                point_id += 1
                yield step
            region_start = region_stop

    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step


class reg3scan(Macro):
    """reg3scan.
    Do an absolute scan of the specified motors with different number of intervals for each region.
    It uses the gscan framework. All the motors will be drived to the same position in each step

    NOTE: Due to a ParamRepeat limitation, integration time has to be
    specified before the regions.
    """

    hints = {'scan' : 'reg3scan'}
    env = ('ActiveMntGrp',)

    param_def = [
        ['motor1',     Type.Moveable, None, 'Motor to move'],
        ['motor2',     Type.Moveable, None, 'Motor to move'],
        ['motor3',     Type.Moveable, None, 'Motor to move'],
        ['integ_time', Type.Float,    None, 'Integration time'],
        ['start_pos',  Type.Float,    None, 'Start position'],
        ['step_region',
         ParamRepeat(['next_pos',  Type.Float,   None, 'next position'],
                     ['region_nr_intervals',  Type.Float,   None, 'Region number of intervals']),
         None, 'List of tuples: (next_pos, region_nr_intervals']
    ]

    def prepare(self, motor1, motor2, motor3, integ_time, start_pos, *regions, **opts):
        self.name='reg3scan'
        self.integ_time = integ_time
        self.start_pos = start_pos
        self.regions = regions
        self.regions_count = len(self.regions)/2

        generator=self._generator
        moveables=[motor1, motor2, motor3]
        env=opts.get('env',{})
        constrains=[]
        self._gScan=SScan(self, generator, moveables, env, constrains)

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time

        point_id = 0
        region_start = self.start_pos
        for r in range(len(self.regions)):
            region_stop, region_nr_intervals = self.regions[r][0], self.regions[r][1]
            positions = numpy.linspace(region_start, region_stop, region_nr_intervals+1)
            if region_start != self.start_pos:
                # positions must be calculated from the start to the end of the region
                # but after the first region, the 'start' point must not be repeated
                positions = positions[1:]
            for p in positions:
                step['positions'] = [p, p, p]
                step['point_id'] = point_id
                point_id += 1
                yield step
            region_start = region_stop

    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step


class a2scan_mod(Macro):
    """a2scan_mod.
    Do an a2scan with the particularity of different intervals per motor: int_mot1, int_mot2.
    If int_mot2 < int_mot1, mot2 will change position every int(int_mot1/int_mot2) steps of mot1.
    It uses the gscan framework.
    """

    hints = {'scan' : 'a2scan_mod'}
    env = ('ActiveMntGrp',)

    param_def = [
       ['motor1',      Type.Moveable, None, 'Motor 1 to move'],
       ['start_pos1',  Type.Float,    None, 'Scan start position 1'],
       ['final_pos1',  Type.Float,    None, 'Scan final position 1'],
       ['nr_interv1',  Type.Integer,  None, 'Number of scan intervals of Motor 1'],
       ['motor2',      Type.Moveable, None, 'Motor 2 to move'],
       ['start_pos2',  Type.Float,    None, 'Scan start position 2'],
       ['final_pos2',  Type.Float,    None, 'Scan final position 2'],
       ['nr_interv2',  Type.Integer,  None, 'Number of scan intervals of Motor 2'],
       ['integ_time',  Type.Float,    None, 'Integration time']
    ]
    
    def prepare(self, motor1, start_pos1, final_pos1, nr_interv1, motor2, start_pos2, final_pos2, nr_interv2, integ_time,
                **opts):
        self.name='a2scan_mod'
        self.integ_time = integ_time
        self.start_pos1 = start_pos1
        self.final_pos1 = final_pos1
        self.nr_interv1 = nr_interv1
        self.start_pos2 = start_pos2
        self.final_pos2 = final_pos2
        self.nr_interv2 = nr_interv2
        
        generator = self._generator
        moveables = [motor1, motor2]
        env = opts.get('env',{})
        constraints = []
        self._gScan=SScan(self, generator, moveables, env, constraints)

    def _generator(self):
        step = {}
        step["integ_time"] =  self.integ_time

        start1, end1, interv1 = self.start_pos1, self.final_pos1, self.nr_interv1
        start2, end2, interv2 = self.start_pos2, self.final_pos2, self.nr_interv2
        
        # Prepare the positions
        positions_m1 = numpy.linspace(start1, end1, interv1+1)
        positions_m2 = numpy.linspace(start2, end2, interv2+1)

        if interv1 > interv2:
            positions_m2 = start2+(float(end2-start2)/interv2)*(numpy.arange(interv1+1)//(float(interv1)/float(interv2)))
        elif interv2 > interv1:
            positions_m1 = start1+(float(end1-start1)/interv1)*(numpy.arange(interv2+1)//(float(interv2)/float(interv1)))

        point_id = 0
        for pos1,pos2 in zip(positions_m1,positions_m2):
            step['point_id'] = point_id
            step['positions'] = [pos1, pos2]
            yield step
            point_id += 1
 
    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step


class ascanc_demo(Macro):
    """
    This is a basic reimplementation of the ascanc` macro for demonstration
    purposes of the Generic Scan framework. The "real" implementation of
    :class:`sardana.macroserver.macros.ascanc` derives from
    :class:`sardana.macroserver.macros.aNscan` and provides some extra features.
    """

    hints = { 'scan' : 'ascanc_demo'} #this is used to indicate other codes that the macro is a scan
    env = ('ActiveMntGrp',) #this hints that the macro requires the ActiveMntGrp environment variable to be set

    param_def = [
       ['motor',      Type.Moveable, None, 'Motor to move'],
       ['start_pos',  Type.Float,    None, 'Scan start position'],
       ['final_pos',  Type.Float,    None, 'Scan final position'],
       ['integ_time', Type.Float,    None, 'Integration time']
    ]

    def prepare(self, motor, start_pos, final_pos, integ_time, **opts):
        self.name='ascanc_demo'
        #parse the user parameters
        self.start = numpy.array([start_pos], dtype='d')
        self.final = numpy.array([final_pos], dtype='d')
        self.integ_time = integ_time
        env = opts.get('env',{}) #the "env" dictionary may be passed as an option
        
        #create an instance of GScan (in this case, of its child, CScan
        self._gScan = CScan(self, 
                            waypointGenerator=self._waypoint_generator, 
                            periodGenerator=self._period_generator, 
                            moveables=[motor], 
                            env=env)
        
    def _waypoint_generator(self):
        #a very simple waypoint generator! only start and stop points!
        yield {"positions":self.start, "waypoint_id": 0}
        yield {"positions":self.final, "waypoint_id": 1}
        

    def _period_generator(self):
        step = {}
        step["integ_time"] =  self.integ_time
        point_no = 0
        while(True): #infinite generator. The acquisition loop is started/stopped at begin and end of each waypoint 
            point_no += 1
            step["point_id"] = point_no
            yield step
    
    def run(self,*args):
        for step in self._gScan.step_scan():
            yield step



class ascan_with_addcustomdata(ascan_demo):
    '''
    example of an ascan-like macro where we demonstrate how to pass custom data to the data handler.
    This is an extension of the ascan_demo macro. Wemake several calls to `:meth:DataHandler.addCustomData`
    exemplifying different features.
    At least the following recorders will act on custom data: 
      - OutputRecorder (this will ignore array data)
      - NXscan_FileRecorder
      - SPEC_FileRecorder (this will ignore array data)
    '''
    def run(self, motor, start_pos, final_pos, nr_interv, integ_time, **opts):
        #we get the datahandler
        dh = self._gScan._data_handler
        #at this point the entry name is not yet set, so we give it explicitly (otherwise it would default to "entry")
        dh.addCustomData('Hello world1', 'dummyChar1', nxpath='/custom_entry:NXentry/customdata:NXcollection')
        #this is the normal scan loop
        for step in self._gScan.step_scan():
            yield step
        #the entry number is known and the default nxpath is used "/<currententry>/custom_data") if none given
        dh.addCustomData('Hello world1', 'dummyChar1')
        #you can pass arrays (but not all recorders will handle them) 
        dh.addCustomData(range(10), 'dummyArray1') 
        #you can pass a custom nxpath *relative* to the current entry
        dh.addCustomData('Hello world2', 'dummyChar2', nxpath='sample:NXsample') 
        
        #calculate a linear fit to the timestamps VS motor positions and store it 
        x = [r.data [motor.getName()] for r in self.data.records]
        y = [r.data['timestamp'] for r in self.data.records]
        fitted_y = numpy.polyval(numpy.polyfit(x,y,1), x)
        dh.addCustomData(fitted_y, 'fittedtime', nxpath='measurement:NXcollection')
        
        #as a bonus, plot the fit
        self.pyplot.plot(x, y, 'ro')
        self.pyplot.plot(x, fitted_y, 'b-')

    